#!/usr/bin/env python


import os
import logging
import shutil
from texttestlib.default import fpdiff
from texttestlib import plugins
from optparse import OptionParser
from io import StringIO


class Filtering(plugins.TestState):
    def __init__(self, name, **kw):
        plugins.TestState.__init__(self, name, briefText="", **kw)

# Generic base class for filtering standard and temporary files


class FilterAction(plugins.Action):
    def __init__(self, useFilteringStates=False):
        self.diag = logging.getLogger("Filter Actions")
        self.useFilteringStates = useFilteringStates

    def __call__(self, test):
        if self.useFilteringStates:
            self.changeToFilteringState(test)

        for fileName, postfix in self.filesToFilter(test):
            self.diag.info("Considering for filtering : " + fileName)
            stem = self.getStem(fileName)
            newFileName = test.makeTmpFileName(stem + "." + test.app.name + postfix, forFramework=1)
            self.performAllFilterings(test, stem, fileName, newFileName)

    def getStem(self, fileName):
        return os.path.basename(fileName).split(".")[0]

    def changeToFilteringState(self, *args):  # pragma: no cover - documentation only
        pass

    def performAllFilterings(self, test, stem, fileName, newFileName):
        currFileName = fileName
        filters = self.makeAllFilters(test, stem, test.app)
        for fileFilter in filters:
            writeFileName = newFileName + "." + fileFilter.postfix
            self.diag.info("Applying " + fileFilter.__class__.__name__ +
                           " to make\n" + writeFileName + " from\n " + currFileName)
            if os.path.isfile(writeFileName):
                self.diag.info("Removing previous file at " + writeFileName)
                os.remove(writeFileName)
            currFile = open(currFileName, errors="ignore")
            writeFile = plugins.openForWrite(writeFileName)
            fileFilter.filterFile(currFile, writeFile)
            writeFile.close()
            currFileName = writeFileName
        if len(filters) > 0 and currFileName != newFileName:
            shutil.move(currFileName, newFileName)

    def getAllFilters(self, test, fileName, app):
        stem = self.getStem(fileName)
        return self.makeAllFilters(test, stem, app)

    def getFilteredText(self, test, fileName, app):
        filters = self.getAllFilters(test, fileName, app)
        inFile = open(fileName, errors="ignore")
        if len(filters) == 0:
            try:
                return inFile.read()
            finally:
                inFile.close()
        for fileFilter in filters:
            self.diag.info("Applying " + fileFilter.__class__.__name__ + " to " + fileName)
            outFile = StringIO()
            fileFilter.filterFile(inFile, outFile)
            inFile.close()
            inFile = outFile
            inFile.seek(0)
        value = outFile.getvalue()
        outFile.close()
        return value

    def makeAllFilters(self, test, stem, app):
        filters = self._makeAllFilters(test, stem, app)
        if len(filters) == 0 and self.changedOs(app):
            return [RunDependentTextFilter([], "")]
        else:
            return filters

    def _makeAllFilters(self, test, stem, app):
        filters = []
        configObj = test
        if test.app is not app:  # happens when testing filtering in the static GUI
            configObj = app

        for filterClass in [RunDependentTextFilter, UnorderedTextFilter]:
            texts = configObj.getCompositeConfigValue(filterClass.configKey, stem)
            if texts:
                filters.append(filterClass(texts, test.getRelPath()))

        return filters

    def changedOs(self, app):
        homeOs = app.getConfigValue("home_operating_system")
        return homeOs != "any" and os.name != homeOs

    def constantPostfix(self, files, postfix):
        return [(file, postfix) for file in files]


class FilterOriginal(FilterAction):
    def filesToFilter(self, test):
        resultFiles, defFiles = test.listApprovedFiles(allVersions=False, defFileCategory="regenerate")
        return self.constantPostfix(resultFiles + defFiles, "origcmp")

    def changeToFilteringState(self, test):
        # Notifications of current status are only useful when doing normal filtering in the GUI
        execMachines = test.state.executionHosts
        freeText = "Filtering stored result files on " + ",".join(execMachines)
        test.changeState(Filtering("initial_filter", executionHosts=execMachines,
                                   freeText=freeText, lifecycleChange="start initial filtering"))


class FilterOnTempFile(FilterAction):
    def _makeAllFilters(self, test, stem, app):
        filters = FilterAction._makeAllFilters(self, test, stem, app)
        floatTolerance = test.getCompositeConfigValue("floating_point_tolerance", stem)
        relTolerance = test.getCompositeConfigValue("relative_float_tolerance", stem)
        floatSplit = test.getCompositeConfigValue("floating_point_split", stem)
        if floatTolerance or relTolerance:
            origFile = test.makeTmpFileName(stem + "." + app.name + "origcmp", forFramework=1)
            if not os.path.isfile(origFile):
                origFile = test.getFileName(stem)
            if origFile and os.path.isfile(origFile):
                filters.append(FloatingPointFilter(origFile, floatTolerance, relTolerance, floatSplit))
        return filters


class FilterTemporary(FilterOnTempFile):
    def filesToFilter(self, test):
        return self.constantPostfix(test.listTmpFiles(), "cmp")

    def changeToFilteringState(self, test):
        # Notifications of current status are only useful when doing normal filtering in the GUI
        execMachines = test.state.executionHosts
        freeText = "Filtering and comparing newly generated result files on " + ",".join(execMachines)
        newState = Filtering("final_filter", executionHosts=execMachines, started=1,
                             freeText=freeText, lifecycleChange="start final filtering and comparison")
        if test.state.category == "killed":
            newState.failedPrediction = test.state
        test.changeState(newState)


class FilterOriginalForScript(FilterOriginal):
    def _makeAllFilters(self, *args):
        return []


class FilterErrorText(FilterAction):
    def _makeAllFilters(self, test, stem, app):
        texts = app.getConfigValue("suppress_stderr_text")
        return [RunDependentTextFilter(texts)]


class FilterProgressRecompute(FilterOnTempFile):
    def filesToFilter(self, test):
        return self.constantPostfix(test.listTmpFiles(), "partcmp")


class FilterResultRecompute(FilterOnTempFile):
    def filesToFilter(self, test):
        result = []
        for fileComp in test.state.allResults:
            # Either of these files might have disappeared
            if fileComp.stdFile and os.path.isfile(fileComp.stdFile):
                result.append((fileComp.stdFile, "origcmp"))
            if fileComp.tmpFile and os.path.isfile(fileComp.tmpFile):
                result.append((fileComp.tmpFile, "cmp"))
        return result


class FloatingPointFilter:
    postfix = "fpdiff"

    def __init__(self, origFileName, tolerance, relative, split):
        self.origFileName = origFileName
        self.tolerance = tolerance if tolerance else None
        self.relative = relative if relative else None
        self.split = split if split else None

    def filterFile(self, inFile, writeFile):
        fromlines = open(self.origFileName, errors="ignore").readlines()
        tolines = inFile.readlines()
        fpdiff.fpfilter(fromlines, tolines, writeFile, self.tolerance, self.relative, split=self.split)


class RunDependentTextFilter(plugins.Observable):
    configKey = "run_dependent_text"
    postfix = "normal"

    def __init__(self, filterTexts, testId=""):
        plugins.Observable.__init__(self)
        self.diag = logging.getLogger("Run Dependent Text")
        self.lineFilters = [LineFilter(text, testId, self.diag) for text in filterTexts]

    def findRelevantFilters(self, file):
        relevantFilters, sectionFilters = [], []
        for lineFilter in self.lineFilters:
            if lineFilter.untrigger is not None:
                sectionFilters.append(lineFilter)
            else:
                relevantFilters.append(lineFilter)
        if sectionFilters:
            # Must preserve the original order
            relevantSectionFilters = self.findRelevantSectionFilters(sectionFilters, file)
            orderedRelevantFilters = []
            for lineFilter in self.lineFilters:
                if lineFilter in relevantFilters:
                    orderedRelevantFilters.append((lineFilter, None))
                else:
                    lastLine = self.getLastLine(relevantSectionFilters, lineFilter)
                    if lastLine:
                        orderedRelevantFilters.append((lineFilter, lastLine))
            return orderedRelevantFilters
        else:
            return [(f, None) for f in relevantFilters]

    def getLastLine(self, filters, lineFilter):
        for f, lastLine in reversed(filters):
            if f is lineFilter:
                return lastLine

    def findRelevantSectionFilters(self, sectionFilters, file):
        lineNumber = 0
        matchedFirst, relevantFilters = [], []
        for line in file:
            lineNumber += 1
            for sectionFilter in matchedFirst:
                if sectionFilter not in relevantFilters and sectionFilter.untrigger.matches(line, lineNumber):
                    relevantFilters.append((sectionFilter, lineNumber))
            for sectionFilter in sectionFilters:
                if sectionFilter not in matchedFirst and \
                        sectionFilter.trigger.matches(line, lineNumber) and not sectionFilter.untrigger.matches(line, lineNumber):
                    matchedFirst.append(sectionFilter)
        for sectionFilter in sectionFilters:
            sectionFilter.trigger.reset()
            sectionFilter.untrigger.reset()
        file.seek(0)
        return relevantFilters

    def filterFile(self, file, newFile, filteredAway=None):
        lineNumber = 0
        seekPoints = []
        lineFilters = self.findRelevantFilters(file)
        for line in file:
            # We don't want to stack up ActionProgreess calls in ThreaderNotificationHandler ...
            self.notifyIfMainThread("ActionProgress")
            lineNumber += 1
            lineFilter, filteredLine, removeCount = self.getFilteredLine(line, lineNumber, lineFilters)
            if removeCount:
                seekPoint = seekPoints[-removeCount - 1]
                self.diag.info("Removing " + repr(removeCount) + " lines")
                newFile.seek(seekPoint)
                newFile.truncate()
                seekPoints = []
            if filteredLine:
                newFile.write(filteredLine)
            else:
                if filteredAway is not None and lineFilter is not None:
                    filteredAway.setdefault(lineFilter, []).append(line)
            seekPoints.append(newFile.tell())

    def getFilteredLine(self, line, lineNumber, lineFilters):
        appliedLineFilter = None
        filteredLine = line
        linesToRemove = 0
        filtersToRemove = []
        alreadyFilteredAway = False
        for lineFilter, lastRelevantLine in lineFilters:
            changed, currFilteredLine, removeCount = lineFilter.applyTo(line, lineNumber, alreadyFilteredAway)
            if lastRelevantLine is not None and lineNumber >= lastRelevantLine:
                filtersToRemove.append((lineFilter, lastRelevantLine))

            if changed:
                appliedLineFilter = lineFilter
                linesToRemove = max(removeCount, linesToRemove)
                if currFilteredLine and filteredLine:
                    line = currFilteredLine
                if filteredLine:
                    filteredLine = currFilteredLine
                if currFilteredLine is None:
                    alreadyFilteredAway = True
        for lineFilter, lastRelevantLine in filtersToRemove:
            lineFilters.remove((lineFilter, lastRelevantLine))

        return appliedLineFilter, filteredLine, linesToRemove


class UnorderedTextFilter(RunDependentTextFilter):
    configKey = "unordered_text"
    postfix = "sorted"

    def filterFile(self, file, newFile):
        unorderedLines = {}
        RunDependentTextFilter.filterFile(self, file, newFile, unorderedLines)
        self.writeUnorderedText(newFile, unorderedLines)

    def writeUnorderedText(self, newFile, lines):
        for filter in self.lineFilters:
            unordered = lines.get(filter, [])
            if len(unordered) == 0:
                continue
            unordered.sort()
            newFile.write("-- Unordered text as found by filter '" + filter.originalText + "' --" + "\n")
            for line in unordered:
                newFile.write(line)
            newFile.write("\n")


class LineNumberTrigger:
    def __init__(self, lineNumber):
        self.lineNumber = lineNumber

    def __repr__(self):
        return "Line number trigger for line " + str(self.lineNumber)

    def matches(self, lineArg, lineNumber):
        return lineNumber == self.lineNumber

    def replace(self, lineArg, newText):
        return newText

    def reset(self):
        pass


class MatchNumberTrigger(plugins.TextTrigger):
    def __init__(self, text, matchNumber):
        self.matchNumber = matchNumber
        self.matchCounter = 0
        plugins.TextTrigger.__init__(self, text)

    def __repr__(self):
        return "Match number trigger for the " + str(self.matchNumber) + ":th match"

    def matches(self, line, *args):
        if plugins.TextTrigger.matches(self, line):
            self.matchCounter += 1
            return self.matchNumber == self.matchCounter
        return False

    def reset(self):
        self.matchCounter = 0


def getWriteDirRegexp(testId):
    testId = testId.replace("\\", "/")
    for char in "+^":
        testId = testId.replace(char, "\\" + char)
    # Some stuff, a date, and the testId (ignore the appId as we don't know when or where)
    # Doesn't handle paths with spaces, which seems hard, but does hardcode the default location of $HOME on Windows...
    posixVersion = '([A-Za-z]:/Documents and Settings)?[^ \'"=]*/[^ "=]*[0-3][0-9][A-Za-z][a-z][a-z][0-9]{6}[^ "=]*/' + testId
    return posixVersion.replace("/", "[/\\\\]+")


class LineFilter:
    dividers = ["{->}", "{[->]}", "{[->}", "{->]}"]
    # All syntax that affects how a match is found
    matcherStrings = ["{LINE ", "{INTERNAL ", "{MATCH "]
    # All syntax that affects what is done when a match is found
    matchModifierStrings = ["{WORD ", "{REPLACE ", "{LINES ", "{PREVLINES "]
    internalExpressions = {"writedir": getWriteDirRegexp}

    def __init__(self, text, testId, diag):
        self.originalText = text
        self.testId = testId
        self.diag = diag
        self.trigger = None
        self.untrigger = None
        self.linesToRemove = 1
        self.prevLinesToRemove = 0
        self.autoRemove = 0
        self.wordNumber = None
        self.replaceText = None
        self.divider = None
        self.removeWordsAfter = False
        self.parseOriginalText()
        self.diag.info("Created trigger : " + repr(self.trigger))

    def makeNew(self, newText):
        return LineFilter(newText, self.testId, self.diag)

    def getInternalExpression(self, parameter):
        method = self.internalExpressions.get(parameter)
        return method(self.testId)

    def makeRegexTrigger(self, parameter):
        expression = self.getInternalExpression(parameter)
        return plugins.TextTrigger(expression)

    def parseOriginalText(self):
        for divider in self.dividers:
            dividerPoint = self.originalText.find(divider)
            if dividerPoint != -1:
                beforeText, afterText = self.originalText.split(divider)
                self.divider = divider
                self.trigger = self.parseText(beforeText)
                self.untrigger = self.parseText(afterText)
                return
        self.trigger = self.parseText(self.originalText)

    def parseText(self, text):
        for matchModifierString in self.matchModifierStrings:
            linePoint = text.find(matchModifierString)
            if linePoint != -1:
                beforeText, afterText, parameter = self.extractParameter(text, linePoint, matchModifierString)
                self.readMatchModifier(matchModifierString, parameter)
                text = beforeText + afterText
        matcherString, text, parameter = self.findMatcherInfo(text)
        return self.createTrigger(matcherString, text, parameter)

    def findMatcherInfo(self, text):
        for matcherString in self.matcherStrings:
            linePoint = text.find(matcherString)
            if linePoint != -1:
                beforeText, afterText, parameter = self.extractParameter(text, linePoint, matcherString)
                return matcherString, beforeText + afterText, parameter
        return "", text, None

    def extractParameter(self, textToParse, linePoint, syntaxString):
        beforeText = textToParse[:linePoint]
        afterText = textToParse[linePoint + len(syntaxString):]
        endPos = afterText.find("}")
        parameter = afterText[:endPos]
        afterText = afterText[endPos + 1:]
        return beforeText, afterText, parameter

    def readMatchModifier(self, matchModifierString, parameter):
        if matchModifierString == "{REPLACE ":
            self.replaceText = parameter
        elif matchModifierString == "{WORD ":
            if parameter.endswith("+"):
                self.removeWordsAfter = True
                self.wordNumber = int(parameter[:-1])
            else:
                self.wordNumber = int(parameter)
            # Somewhat non-intuitive to count from 0...
            if self.wordNumber > 0:
                self.wordNumber -= 1
        elif matchModifierString == "{LINES ":
            self.linesToRemove = int(parameter)
        elif matchModifierString == "{PREVLINES ":
            self.prevLinesToRemove = int(parameter)

    def createTrigger(self, matcherString, text, parameter):
        if matcherString == "{LINE ":
            return LineNumberTrigger(int(parameter))
        elif matcherString == "{INTERNAL " and parameter in self.internalExpressions:
            return self.makeRegexTrigger(parameter)
        elif matcherString == "{MATCH ":
            return MatchNumberTrigger(text, int(parameter))
        else:
            return plugins.TextTrigger(text)

    def isMultiLine(self):
        return self.linesToRemove > 1 or self.prevLinesToRemove > 0 or self.untrigger is not None

    def applyTo(self, line, lineNumber=0, alreadyFilteredAway=False):
        if self.autoRemove:
            return self.applyAutoRemove(line)
        elif alreadyFilteredAway and not self.isMultiLine():
            return False, None, 0

        if self.trigger.matches(line, lineNumber):
            self.diag.info(repr(self.trigger) + " matched " + line.rstrip())
            return self.applyMatchingTrigger(line)
        else:
            return False, line, 0

    def applyAutoRemove(self, line):
        if self.untrigger:
            if self.untrigger.matches(line.rstrip()):
                self.diag.info(repr(self.untrigger) + " (end) matched " + line.rstrip())
                self.autoRemove = 0
                if self.divider.endswith("]}"):
                    return True, None, 0
                else:
                    return False, line, 0
        else:
            self.autoRemove -= 1
        return True, self.filterWords(line), 0

    def applyMatchingTrigger(self, line):
        if self.untrigger:
            self.autoRemove = 1
            return self.divider.startswith("{["), self.filterWords(line), 0
        if self.linesToRemove:
            self.autoRemove = self.linesToRemove - 1
        return True, self.filterWords(line, self.trigger), self.prevLinesToRemove

    def filterWords(self, line, trigger=None):
        if self.wordNumber != None:
            stripped = line.rstrip()
            postfix = line.replace(stripped, "", 1)
            words = stripped.split(" ")
            self.diag.info("Removing word " + str(self.wordNumber) + " from " + repr(words))
            realNumber = self.findRealWordNumber(words)
            self.diag.info("Real number was " + str(realNumber))
            if realNumber < len(words):
                if self.removeWordsAfter:
                    words = words[:realNumber]
                    if self.replaceText:
                        words.append(self.replaceText)
                else:
                    if self.replaceText:
                        words[realNumber] = self.replaceText
                    else:
                        del words[realNumber]
            if realNumber == -1 or realNumber >= len(words) - 1:  # Trim trailing spaces for words at end or beyond
                postfix = "\n"
            return " ".join(words).rstrip() + postfix
        elif trigger and self.replaceText != None:
            return trigger.replace(line.rstrip("\n"), self.replaceText) + "\n"

    def findRealWordNumber(self, words):
        if self.wordNumber < 0:
            return self.findRealWordNumberBackwards(words)
        wordNumber = 0
        for realWordNumber in range(len(words)):
            if len(words[realWordNumber]):
                if wordNumber == self.wordNumber:
                    return realWordNumber
                wordNumber += 1
        return len(words) + 1

    def findRealWordNumberBackwards(self, words):
        wordNumber = -1
        for index in range(len(words)):
            realWordNumber = -1 - index
            word = words[realWordNumber]
            if len(word):
                if wordNumber == self.wordNumber:
                    return realWordNumber
                wordNumber -= 1
        return len(words) + 1



import os
import shutil
import re
import stat
import subprocess
import glob
import logging
import difflib
import time
import sys
from texttestlib import plugins
from texttestlib.jobprocess import killProcessAndChildren
from .runtest import Killed
from collections import OrderedDict
from string import Template


def getScriptArgs(script):
    args = script.split()
    scriptName = args[0]
    instScript = plugins.installationPath(os.path.join("libexec", scriptName))
    if instScript:
        args = [instScript] + args[1:]

    return args


class MakeWriteDirectory(plugins.Action):
    def __call__(self, test):
        test.makeWriteDirectory()

    def setUpApplication(self, app):
        app.makeWriteDirectory()


class PrepareWriteDirectory(plugins.Action):
    storytextDirsCopied = set()

    def __init__(self, ignoreCatalogues):
        self.diag = logging.getLogger("Prepare Writedir")
        self.ignoreCatalogues = ignoreCatalogues
        self.handledRequiredPaths = set()
        if self.ignoreCatalogues:
            self.diag.info("Ignoring all information in catalogue files")

    def __call__(self, test):
        test.backupTemporaryData()
        machine, remoteTmpDir = test.app.getRemoteTestTmpDir(test)
        if remoteTmpDir:
            test.app.ensureRemoteDirExists(machine, remoteTmpDir)
            remoteCopy = plugins.Callable(self.copyDataRemotely, test, machine, remoteTmpDir)
        else:
            remoteCopy = None

        self.collateAllPaths(test, remoteCopy)
        test.createPropertiesFiles()

    def collateAllPaths(self, test, remoteCopy):
        self.collatePaths(test, "copy_test_path", self.copyTestPath, remoteCopy)
        self.collatePaths(test, "copy_test_path_merge", self.copyTestPath, remoteCopy, mergeData=True)
        self.collatePaths(test, "partial_copy_test_path", self.partialCopyTestPath, remoteCopy)
        self.collatePaths(test, "link_test_path", self.linkTestPath, remoteCopy)

    def collatePaths(self, test, configListName, *args, **kwargs):
        for configName in test.getConfigValue(configListName, expandVars=False):
            self.collatePath(test, configName, *args, **kwargs)

    def handleNoTestData(self, test, configName, sourcePaths):
        if configName in test.getConfigValue("test_data_require", expandVars=False):
            msg = "No data source found for required test data '" + configName + "'"
            if sourcePaths:
                msg += "\nNo such file or directory : " + ", ".join(sourcePaths)
            raise plugins.TextTestError(msg)

    def handleRequiredData(self, test, configName, sourcePaths):
        unseenSource = [p for p in sourcePaths if p not in self.handledRequiredPaths]
        if unseenSource and configName in test.getConfigValue("test_data_require", expandVars=False):
            test.notify("RequiredTestData", sourcePaths)
            self.handledRequiredPaths.update(sourcePaths)

    def collatePath(self, test, configName, collateMethod, remoteCopy, mergeData=False):
        targetPath = self.getTargetPath(test, configName)
        sourceFileName = self.getSourceFileName(configName, test)
        if not targetPath or not sourceFileName:  # Can happen with e.g. empty environment
            self.handleNoTestData(test, configName, [])
            return
        plugins.ensureDirExistsForFile(targetPath)
        sourcePaths = self.getSourcePaths(test, configName, sourceFileName)
        if sourcePaths:
            self.handleRequiredData(test, configName, sourcePaths)
        collated = False
        for sourcePath in self.getSortedSourcePaths(sourcePaths, mergeData):
            if os.path.exists(sourcePath):
                self.diag.info("Collating " + configName + " from " + repr(sourcePath) +
                               "\nto " + repr(targetPath))
                collateMethod(test, sourcePath, targetPath)
                collated = True

        if not collated:
            self.diag.info("No test data present in " + repr(sourcePaths))
            self.handleNoTestData(test, configName, sourcePaths)

        if remoteCopy and targetPath:
            remoteCopy(targetPath)

        envVarToSet = self.findDataEnvironment(test, configName)
        if envVarToSet and targetPath:
            self.diag.info("Setting env. variable " + envVarToSet + " to " + targetPath)
            test.setEnvironment(envVarToSet, targetPath)

    def getSortedSourcePaths(self, sourcePaths, mergeData):
        if mergeData:
            if sourcePaths and os.path.isdir(sourcePaths[-1]):
                # Merging directories copies specific information first and refuses to overwrite it, must switch the order
                return reversed(sourcePaths)
            else:
                # Merging files starts with most general and appends specific ones, keep the order
                return sourcePaths
        else:
            # Don't merge, just use the most specific data
            return sourcePaths[-1:]

    def copyDataRemotely(self, sourcePath, test, machine, remoteTmpDir):
        if os.path.exists(sourcePath):
            copyScript = test.getCompositeConfigValue(
                "copy_test_path_script", os.path.basename(sourcePath), expandVars=False)
            if copyScript:
                scriptSource = os.path.join(remoteTmpDir, "scriptSource")
                test.app.copyFileRemotely(sourcePath, "localhost", scriptSource, machine)
                cmdArgs = getScriptArgs(copyScript) + [scriptSource,
                                                       os.path.join(remoteTmpDir, os.path.basename(sourcePath))]
                test.app.runCommandOn(machine, cmdArgs)
            else:
                test.app.copyFileRemotely(sourcePath, "localhost", remoteTmpDir, machine)

    def getEnvironmentSourcePath(self, configName, test):
        pathName = self.getPathFromEnvironment(configName, test)
        if pathName != configName:
            self.diag.info("Found source file name for " + configName + " = " + repr(pathName))
            return pathName

    def getPathFromEnvironment(self, configName, test):
        expanded = Template(configName).safe_substitute(test.environment)
        if expanded:  # Don't do normpath on empty strings, you get "." which causes trouble later...
            return os.path.normpath(expanded)
        else:
            return expanded

    def getTargetPath(self, test, configName):
        # handle environment variables
        localName = os.path.basename(self.getPathFromEnvironment(configName, test))
        if localName:
            return test.makeTmpFileName(localName, forComparison=0)

    def getSourcePaths(self, test, configName, fileName):
        if os.path.isabs(fileName):
            return [fileName]
        else:
            return test.getAllPathNames(fileName, configName)  # most general first

    def getSourceFileName(self, configName, test):
        # These can refer to environment variables or to paths within the test structure
        if configName.startswith("$"):
            return self.getEnvironmentSourcePath(configName, test)
        else:
            return configName

    def findDataEnvironment(self, test, configName):
        self.diag.info("Finding env. var name from " + configName)
        if configName.startswith("$"):
            return configName[1:]

        envVarDict = test.getConfigValue("test_data_environment")
        return envVarDict.get(configName)

    def copyTestPath(self, test, fullPath, target):
        copyScript = test.getCompositeConfigValue("copy_test_path_script", os.path.basename(target))
        if copyScript:
            try:
                args = getScriptArgs(copyScript)
                subprocess.call(args + [fullPath, target], env=test.getRunEnvironment())
                return
            except OSError:
                pass  # If this doesn't work, assume it's on the remote machine and we'll handle it later

        if os.path.isfile(fullPath):
            if os.path.isfile(target):
                with open(target, "a") as f:
                    f.write(open(fullPath).read())
            else:
                self.copyfile(fullPath, target)
        if os.path.isdir(fullPath):
            self.copytree(fullPath, target)

    def copytimes(self, src, dst):
        if os.path.isdir(src) and os.name == "nt":
            # Windows doesn't let you update modification times of directories!
            return
        # copy modification times, but not permissions. This is a copy of half of shutil.copystat
        st = os.stat(src)
        if hasattr(os, 'utime'):
            os.utime(dst, (st[stat.ST_ATIME], st[stat.ST_MTIME]))

    def copytree(self, src, dst):
        # Code is a copy of shutil.copytree, with copying modification times
        # so that we can tell when things change...
        names = os.listdir(src)
        if not os.path.exists(dst):
            os.mkdir(dst)
        for name in names:
            srcname = os.path.join(src, name)
            dstname = os.path.join(dst, name)
            if os.path.isfile(dstname) or os.path.islink(dstname):
                continue
            try:
                if os.path.islink(srcname):
                    self.copylink(srcname, dstname)
                elif os.path.isdir(srcname):
                    self.copytree(srcname, dstname)
                else:
                    self.copyfile(srcname, dstname)
            except (IOError, os.error) as why:
                print("Can't copy", srcname, "to", dstname, ":", why)
        # Last of all, keep the modification time as it was
        self.copytimes(src, dst)

    def copylink(self, srcname, dstname):
        linkto = srcname
        if os.path.islink(srcname):
            linkto = os.readlink(srcname)
        os.symlink(linkto, dstname)

    def copyfile(self, srcname, dstname):
        # Basic aim is to keep the permission bits and times where possible, but ensure it is writeable
        shutil.copy2(srcname, dstname)
        plugins.makeWriteable(dstname)

    def linkTestPath(self, test, fullPath, target):
        # Linking doesn't exist on windows!
        if os.name != "posix":
            return self.copyTestPath(test, fullPath, target)
        if not os.path.exists(target):
            os.symlink(fullPath, target)
        else:  # pragma : no cover
            raise plugins.TextTestError("File already existed at " + target + "\nTrying to link to " + fullPath)

    def partialCopyTestPath(self, test, sourcePath, targetPath):
        # Linking doesn't exist on windows!
        if os.name != "posix":
            return self.copyTestPath(test, sourcePath, targetPath)
        modifiedPaths = self.getModifiedPaths(test, sourcePath, os.path.basename(targetPath))
        if modifiedPaths is None:
            # If we don't know, assume anything can change...
            self.copyTestPath(test, sourcePath, targetPath)
        elif sourcePath not in modifiedPaths:
            self.linkTestPath(test, sourcePath, targetPath)
        else:
            os.mkdir(targetPath)
            self.diag.info("Copying/linking for Test " + repr(test))
            writeDirs = self.copyAndLink(sourcePath, targetPath, modifiedPaths)
            # Link everywhere new files appear from the write directory for ease of collection
            for writeDir in writeDirs:
                self.diag.info("Creating bypass link to " + writeDir)
                linkTarget = test.makeTmpFileName(os.path.basename(writeDir), forComparison=0)
                if linkTarget != writeDir and not os.path.exists(linkTarget):
                    # Don't link locally - and it's possible to have the same name twice under different paths
                    os.symlink(writeDir, linkTarget)
            self.copytimes(sourcePath, targetPath)

    def copyAndLink(self, sourcePath, targetPath, modifiedPaths):
        writeDirs = []
        self.diag.info("Copying/linking from " + sourcePath)
        modPathsLocal = modifiedPaths[sourcePath]
        self.diag.info("Modified paths here " + repr(modPathsLocal))
        for file in os.listdir(sourcePath):
            sourceFile = os.path.normpath(os.path.join(sourcePath, file))
            targetFile = os.path.join(targetPath, file)
            if sourceFile in modPathsLocal:
                if os.path.isdir(sourceFile):
                    os.mkdir(targetFile)
                    writeDirs += self.copyAndLink(sourceFile, targetFile, modifiedPaths)
                    self.copytimes(sourceFile, targetFile)
                else:
                    self.copyfile(sourceFile, targetFile)
            else:
                self.handleReadOnly(sourceFile, targetFile)
        if self.isWriteDir(targetPath, modPathsLocal):
            self.diag.info("Registering " + targetPath + " as a write directory")
            writeDirs.append(targetPath)
        return writeDirs

    def handleReadOnly(self, sourceFile, targetFile):
        try:
            self.copylink(sourceFile, targetFile)
        except OSError:  # pragma : no cover
            print("Failed to create symlink " + targetFile)

    def isWriteDir(self, dummy, modPaths):
        for modPath in modPaths:
            if not os.path.isdir(modPath):
                return True
        return False

    def getModifiedPaths(self, test, sourcePath, sourceNameInCatalogue):
        catFile = test.getFileName("catalogue")
        if not catFile or self.ignoreCatalogues:
            # This means we don't know
            return None
        # Catalogue file is actually relative to temporary directory, need to take one level above...
        rootDir = os.path.split(sourcePath)[0]
        fullPaths = {rootDir: []}
        currentPaths = [rootDir]
        for line in open(catFile).readlines():
            fileName, indent = self.parseCatalogue(line)
            if not fileName:
                continue

            if fileName == sourceNameInCatalogue and indent == 1:
                fileName = os.path.basename(sourcePath)

            prevPath = currentPaths[indent - 1]
            fullPath = os.path.join(prevPath, fileName)
            if indent >= len(currentPaths):
                currentPaths.append(fullPath)
            else:
                currentPaths[indent] = fullPath
            if fullPath not in fullPaths:
                fullPaths[fullPath] = []
            if not fullPath in fullPaths[prevPath]:
                fullPaths[prevPath].append(fullPath)
        del fullPaths[rootDir]
        return fullPaths

    def parseCatalogue(self, line):
        pos = line.rfind("----")
        if pos == -1:
            return None, None
        pos += 4
        dashes = line[:pos]
        indent = len(dashes) // 4
        fileName = line.strip()[pos:]
        return fileName, indent

    def setUpSuite(self, suite):
        if suite.parent is None:
            self.tryCopySUTRemotely(suite)

    def isCopy(self, app):
        return any(("copy_" in v for v in app.versions))

    def tryCopySUTRemotely(self, suite):
        # Copy the executables remotely, if necessary
        machine, tmpDir = suite.app.getRemoteTmpDirectory()
        if tmpDir:
            if not self.isCopy(suite.app):
                self.copySUTRemotely(machine, tmpDir, suite)
                self.copyStorytextDirRemotely(machine, tmpDir, suite)
            suite.reloadTestConfigurations()

    def tryCopyPathRemotely(self, path, fullTmpDir, machine, app):
        self.diag.info("Trying to copy " + path + " to " + fullTmpDir)
        if os.path.isabs(path) and os.path.exists(path) and not app.pathExistsRemotely(path, machine):
            # If not absolute, assume it's an installed program
            # If it doesn't exist locally, it must already exist remotely or we'd have raised an error by now
            remotePath = os.path.join(fullTmpDir, os.path.basename(path))
            app.copyFileRemotely(path, "localhost", remotePath, machine)
            self.diag.info("Copied " + path + " to " + remotePath)
            return remotePath

    def copySUTFileIfNeeded(self, machine, suite, fullTmpDir, checkout, remoteCheckout, localFile):
        if remoteCheckout and localFile.startswith(checkout):
            return localFile.replace(checkout, remoteCheckout)  # We've copied it already, don't do it again...
        else:
            return self.tryCopyPathRemotely(localFile, fullTmpDir, machine, suite.app)

    def overrideExecutable(self, app, remoteFile):
        self.diag.info("Setting executable to " + repr(remoteFile))
        app.setConfigOverride("executable", remoteFile)

    def copySUTRemotely(self, machine, tmpDir, suite):
        self.diag.info("Copying SUT for " + repr(suite.app) + " to machine " + machine + " at " + tmpDir)
        fullTmpDir = os.path.join(tmpDir, "system_under_test")
        suite.app.ensureRemoteDirExists(machine, fullTmpDir)
        checkout = suite.app.getCheckoutForDisplay()
        remoteCheckout = self.tryCopyPathRemotely(checkout, fullTmpDir, machine, suite.app)
        if remoteCheckout:
            suite.app.checkout = remoteCheckout

        for interpreterName, localFile in list(suite.getConfigValue("interpreters").items()):
            remoteFile = self.copySUTFileIfNeeded(machine, suite, fullTmpDir, checkout, remoteCheckout, localFile)
            if remoteFile:
                self.diag.info("Setting interpreter " + repr(interpreterName) + " to " + repr(remoteFile))
                # For convenience, so we don't have to set it everywhere...
                suite.app.addConfigEntryOverride(interpreterName, remoteFile, "interpreters")

        localFile = suite.getConfigValue("executable")
        remoteFile = self.copySUTFileIfNeeded(machine, suite, fullTmpDir, checkout, remoteCheckout, localFile)
        if remoteFile:
            # For convenience, so we don't have to set it everywhere...
            self.overrideExecutable(suite.app, remoteFile)
        elif not remoteCheckout:
            self.overrideExecutable(suite.app, localFile)  # Expand TEXTTEST_CHECKOUT, TEXTTEST_ROOT etc

        scripts = suite.getConfigValue("copy_test_path_script")
        newScripts = {}
        for fileName, script in list(scripts.items()):
            if script:
                localScript = getScriptArgs(script)[0]
                remoteScript = self.tryCopyPathRemotely(localScript, fullTmpDir, machine, suite.app)
                if remoteScript:
                    self.diag.info("Setting copy_test_path_script for " + repr(fileName) + " to " + repr(remoteFile))
                    newScripts[fileName] = remoteScript
        if newScripts:
            suite.app.setConfigOverride("copy_test_path_script", newScripts)

    def copyStorytextDirRemotely(self, machine, tmpDir, suite):
        storytextDir = suite.getEnvironment("STORYTEXT_HOME_LOCAL")
        if storytextDir:
            newbasename = os.path.basename(storytextDir)
            if os.path.isdir(storytextDir) and storytextDir not in self.storytextDirsCopied:
                self.storytextDirsCopied.add(storytextDir)
                remoteName = os.path.join(tmpDir, newbasename)
                suite.app.copyFileRemotely(storytextDir, "localhost", remoteName, machine)


# Class for automatically adding things to test environment files...
class TestEnvironmentCreator:
    def __init__(self, test, optionMap):
        self.test = test
        self.optionMap = optionMap
        self.diag = logging.getLogger("Environment Creator")

    def getVariables(self, allVars):
        vars, props = [], []
        self.diag.info("Creating environment for " + repr(self.test))
        if self.topLevel():
            vars.append(("TEXTTEST_ROOT", self.test.app.getDirectory())) # Full path to the root directory for each test application
            checkout = self.test.app.checkout
            if checkout:
                vars.append(("TEXTTEST_CHECKOUT", checkout))  # Full path to the checkout directory
            localCheckout = self.test.app.getCheckout()
            if localCheckout and localCheckout != checkout:
                vars.append(("TEXTTEST_CHECKOUT_NAME", localCheckout))  # Local name of the checkout directory
            vars.append(("TEXTTEST_SANDBOX_ROOT", self.test.app.localWriteDirectory)) # Full path to the sandbox root directory
            if self.test.getConfigValue("use_case_record_mode") == "GUI":
                usecaseRecorder = self.test.getConfigValue("use_case_recorder")
                # Mostly to make sure StoryText's own tests have a chance of working
                # Almost any other test suite shouldn't be doing this...
                envSetInTests = list(self.test.getConfigValue("test_data_environment").values())
                if usecaseRecorder != "none":
                    if "STORYTEXT_HOME" not in envSetInTests:
                        if not usecaseRecorder:
                            usecaseRecorder = "ui_simulation"
                        usecaseDir = os.path.join(self.test.app.getDirectory(), usecaseRecorder + "_files")
                        remoteTmpDir = self.test.app.getRemoteTmpDirectory()[1]
                        if remoteTmpDir:
                            # We pretend it's a local temporary file
                            # so that the $HOME reference gets preserved in runtest.py
                            # (no guarantee $HOME is the same remotely)
                            fakeRemotePath = os.path.join(self.test.app.writeDirectory, os.path.basename(usecaseDir))
                            vars.append(("STORYTEXT_HOME", fakeRemotePath))
                            vars.append(("STORYTEXT_HOME_LOCAL", usecaseDir))
                        else:
                            vars.append(("STORYTEXT_HOME", usecaseDir))

                    eclipseHome = self.findValueFromFiles(allVars, "ECLIPSE_HOME")
                    if eclipseHome:
                        # StoryText with RCP. We add the necessary to the classpath...
                        expandedHome = self.getExpanded(allVars + vars, eclipseHome)
                        jars = self.findJars(expandedHome)
                        vars.append(("CLASSPATH", os.pathsep.join(jars)))

                if os.name == "posix":
                    from .virtualdisplay import VirtualDisplayResponder
                    if VirtualDisplayResponder.instance:
                        for var, value in VirtualDisplayResponder.instance.getVariablesToSet():
                            vars.append((var, value))
        elif self.testCase():
            useCaseVars = self.getUseCaseVariables()
            if self.useJavaRecorder():
                props += useCaseVars
            else:
                vars += useCaseVars
            vars += self.getPathVariables()
        return vars, props

    def getExpanded(self, allVars, path):
        # Expand variables recursively!
        d = dict(allVars)
        while True:
            temp = Template(path)
            newPath = temp.safe_substitute(d)
            if newPath == path:
                return newPath
            else:
                path = newPath

    def findJars(self, eclipseHome):
        names = [("org.eclipse.equinox.launcher_", ""), ("org.eclipse.swtbot", ".testscript")]
        if not os.path.isdir(eclipseHome):
            # Assume that eclipse home is required!
            self.test.notify("MissingRequiredTestData", [eclipseHome])
            if not os.path.isdir(eclipseHome):
                return []
        home = os.path.join(eclipseHome, "plugins")
        jars = []
        for name, suffix in names:
            for fileName in sorted(os.listdir(home)):
                if fileName.startswith(name) and suffix in fileName:
                    jars.append(os.path.join(home, fileName))
                    break

        return jars

    def findValueFromFiles(self, allVars, varName):
        # allVars has the most general first
        # If the variable is overridden it will appear several times in the list,
        # so we start at the end
        for var, value in reversed(allVars):
            if var == varName:
                return value

    def topLevel(self):
        return self.test.parent is None

    def testCase(self):
        return self.test.classId() == "test-case"

    def getUseCaseVariables(self):
        # Here we assume the application uses either StoryText or JUseCase
        # StoryText reads environment variables, but you can't do that from java,
        # so we have a "properties file" set up as well. Do both always, to save forcing
        # apps to tell us which to do...
        usecaseFile = self.test.getFileName("usecase")
        replayUseCase = self.findReplayUseCase(usecaseFile)
        vars = []
        if replayUseCase is not None:
            vars.append(self.getReplayScriptVariable(replayUseCase))
            delay = self.optionMap.get("delay", 0)
            if float(delay) > 0:
                vars.append(self.getReplayDelayVariable(delay))

            screenshot = "screenshot" in self.optionMap
            if screenshot:
                vars.append(("USECASE_REPLAY_SCREENSHOTS", "1")) # Whether to take screenshots between each action in GUI tests
        if usecaseFile or self.isRecording():
            # Re-record if recorded files are already present or recording explicitly requested
            vars.append(self.getRecordScriptVariable(self.test.makeTmpFileName("usecase")))
        return vars

    def isRecording(self):
        return "record" in self.optionMap

    def findReplayUseCase(self, usecaseFile):
        if not self.isRecording():
            if usecaseFile:
                return self.copyRemoteReplayFilesIfNeeded(usecaseFile)
            elif "USECASE_REPLAY_SCRIPT" in os.environ and not self.useJavaRecorder():
                return ""  # Clear our own script, if any, for further apps wanting to use StoryText

    def copyRemoteReplayFilesIfNeeded(self, path):
        machine, remoteTmpDir = self.test.app.getRemoteTestTmpDir(self.test)
        if remoteTmpDir:
            newbasename = "replay_usecase"
            # Sometimes more than one UI is run, copy all files with 'usecase' in the name
            usecasePattern = os.path.join(os.path.dirname(path), "*usecase*." + self.test.app.name)
            for currPath in sorted(glob.glob(usecasePattern)):
                currStem = os.path.basename(currPath).split(".")[0]
                remoteName = os.path.join(remoteTmpDir, currStem.replace("usecase", newbasename))
                self.test.app.copyFileRemotely(currPath, "localhost", remoteName, machine)
            # Don't return remoteName, we pretend it's a local temporary file
            # so that the $HOME reference gets preserved (no guarantee $HOME is the same remotely)
            return os.path.join(self.test.getDirectory(temporary=1), newbasename)
        else:
            return path

    def useJavaRecorder(self):
        return self.test.getConfigValue("use_case_recorder") == "jusecase"

    def getReplayScriptVariable(self, replayScript):
        if self.useJavaRecorder():
            return "replay", replayScript, "jusecase"
        else:
            return "USECASE_REPLAY_SCRIPT", replayScript  # Full path to the script to replay in GUI tests

    def getReplayDelayVariable(self, delay):
        if self.useJavaRecorder():
            return "delay", delay, "jusecase"
        else:
            return "USECASE_REPLAY_DELAY", delay  # Time to wait between each action in GUI tests

    def getRecordScriptVariable(self, recordScript):
        self.diag.info("Enabling recording")
        if self.useJavaRecorder():
            return "record", recordScript, "jusecase"
        else:
            return "USECASE_RECORD_SCRIPT", recordScript  # Full path to the script to record in GUI tests

    def getPathVariables(self):
        testDir = self.test.getDirectory(temporary=1)
        localTestDir = self.test.getDirectory(temporary=1, local=1)
        vars = [("TEXTTEST_SANDBOX", localTestDir),  # Full path to the sandbox directory
                ("TEXTTEST_LOG_DIR", testDir)]  # Full path to where TextTest will write the SUT's log files
        # Always include the working directory of the test in PATH, to pick up fake
        # executables provided as test data. Allow for later expansion...
        for pathVar in self.getPathVars():
            newPathVal = localTestDir + os.pathsep + "$" + pathVar
            vars.append((pathVar, newPathVal))
        return vars

    def getPathVars(self):
        pathVars = ["PATH"]
        for dataFile in self.test.getDataFileNames():
            if dataFile.endswith(".py") and "PYTHONPATH" not in pathVars:
                pathVars.append("PYTHONPATH")
            elif (dataFile.endswith(".jar") or dataFile.endswith(".class")) and "CLASSPATH" not in pathVars:
                pathVars.append("CLASSPATH")
        return pathVars


class CollateFiles(plugins.Action):
    def __init__(self):
        self.filesPresentBefore = {}
        self.collationProc = None
        self.diag = logging.getLogger("Collate Files")

    def expandCollations(self, test):
        newColl = OrderedDict()
        coll = test.getConfigValue("collate_file")
        self.diag.info("coll initial:" + str(coll))
        for targetPattern in sorted(coll.keys()):
            sourcePatterns = coll.get(targetPattern)
            if not glob.has_magic(targetPattern):
                newColl[targetPattern] = sourcePatterns
                continue

            # add each file to newColl by transferring wildcards across
            for sourcePattern in sourcePatterns:
                testDir, sourcePaths = self.findPaths(test, sourcePattern)
                for sourcePath in sourcePaths:
                    # Use relative paths: easier to debug and SequenceMatcher breaks down if strings are longer than 200 chars
                    relativeSourcePath = plugins.relpath(sourcePath, testDir)
                    newTargetStem = self.makeTargetStem(targetPattern, sourcePattern, relativeSourcePath)
                    self.diag.info("New collation to " + newTargetStem + " : from " + relativeSourcePath)
                    newColl.setdefault(newTargetStem, []).append(sourcePath)
        return list(newColl.items())

    def makeTargetStem(self, targetPattern, sourcePattern, sourcePath):
        newTargetStem = targetPattern
        for wildcardMatch in self.findWildCardMatches(sourcePattern, sourcePath):
            newTargetStem = newTargetStem.replace("*", wildcardMatch, 1)
        return newTargetStem.replace("*", "WILDCARD").replace(".", "_")

    def inSquareBrackets(self, pattern, pos):
        closeBracket = pattern.find("]", pos)
        if closeBracket == -1:
            return False
        openBracket = pattern.find("[", pos)
        return openBracket == -1 or closeBracket < openBracket

    def findWildCardMatches(self, pattern, result):
        patternSep = os.sep if os.sep in pattern else "/"
        parts = list(zip(pattern.split(patternSep), result.split(os.sep)))
        allMatches = []
        # We take wildcards in the file name first, then those in directory names
        for subPattern, subResult in reversed(parts):
            allMatches += self._findWildCardMatches(subPattern, subResult)
        return allMatches

    def _findWildCardMatches(self, pattern, result):
        matcher = difflib.SequenceMatcher(None, pattern, result)
        wildcardStart = 0
        matches = []
        self.diag.info("Trying to find wildcard matches in " + repr(result) + " from " + repr(pattern))
        for patternPos, resultPos, length in matcher.get_matching_blocks():
            if length == 0 or self.inSquareBrackets(pattern, patternPos):
                continue
            self.diag.info("Found match of length " + repr(length) + " at positions " + repr((patternPos, resultPos)))
            if resultPos or patternPos:
                matches.append(result[wildcardStart:resultPos])
            wildcardStart = resultPos + length
        remainder = result[wildcardStart:]
        if remainder:
            matches.append(remainder)
        self.diag.info("Found matches : " + repr(matches))
        return matches

    def __call__(self, test):
        if test not in self.filesPresentBefore:
            self.filesPresentBefore[test] = self.getFilesPresent(test)
        else:
            self.tryFetchRemoteFiles(test)
            self.collate(test)
            self.removeUnwanted(test)

    def containsRegexps(self, filePath, regexps):
        with open(filePath) as f:
            for line in f:
                if any((r.search(line) for r in regexps)):
                    return True
        return False

    def removeUnwantedFile(self, filePath):
        self.diag.info("Trying to remove generated file " + os.path.basename(filePath))
        try:
            # Checking for existence too dependent on file server (?)
            os.remove(filePath)
        except EnvironmentError:
            pass

    def removeUnwanted(self, test):
        for stem in test.getConfigValue("discard_file"):
            filePath = test.makeTmpFileName(stem)
            self.removeUnwantedFile(filePath)

        for stemPattern, texts in list(test.getConfigValue("discard_file_text").items()):
            if not texts:
                continue
            if stemPattern == "default":
                stemPattern = "*"
            regexps = list(map(re.compile, texts))
            for filePath in glob.glob(test.makeTmpFileName(stemPattern)):
                if self.containsRegexps(filePath, regexps):
                    self.removeUnwantedFile(filePath)

    def findEditedFiles(self, test, patterns):
        editedFiles = []
        for pattern in patterns:
            for fullpath in self.findPaths(test, pattern)[1]:
                if self.testEdited(test, fullpath):
                    editedFiles.append(fullpath)
                else:
                    self.diag.info("Found " + fullpath + " but it wasn't edited")
        return editedFiles

    def collate(self, test):
        for targetStem, sourcePatterns in self.expandCollations(test):
            sourceFiles = self.findEditedFiles(test, sourcePatterns)
            if sourceFiles:
                targetFile = test.makeTmpFileName(targetStem)
                collationErrFile = test.makeTmpFileName(targetStem + ".collate_errs", forFramework=1)
                self.diag.info("Extracting " + ",".join(sourceFiles) + " to " + targetFile)
                self.extract(test, sourceFiles, targetFile, collationErrFile)

    def tryFetchRemoteFiles(self, test):
        machine, remoteTmpDir = test.app.getRemoteTestTmpDir(test)
        if remoteTmpDir:
            self.fetchRemoteFiles(test, machine, remoteTmpDir)

    def fetchRemoteFiles(self, test, machine, tmpDir):
        sourcePaths = os.path.join(plugins.quote(tmpDir), "*")
        test.app.copyFileRemotely(sourcePaths, machine, test.getDirectory(temporary=1), "localhost")

    def getFilesPresent(self, test):
        files = OrderedDict()
        for sourcePatterns in list(test.getConfigValue("collate_file").values()):
            for sourcePattern in sourcePatterns:
                for fullPath in self.findPaths(test, sourcePattern)[1]:
                    self.diag.info("Pre-existing file found " + fullPath)
                    files[fullPath] = plugins.modifiedTime(fullPath)
        return files

    def testEdited(self, test, fullPath):
        filesBefore = self.filesPresentBefore[test]
        if fullPath not in filesBefore:
            return True
        return filesBefore[fullPath] != plugins.modifiedTime(fullPath)

    def alreadyCollated(self, test, path, sourcePattern):
        if "/" not in sourcePattern:
            parts = os.path.basename(path).split(".")
            if len(parts) > 1 and parts[1] == test.app.name:
                return True  # Don't collate generated files
        return False

    def glob(self, test, sourcePattern):
        # Test name may contain glob meta-characters, and there is no way to quote them (see comment in fnmatch.py)
        # So we can't just form an absolute path and glob that
        localTestDir = test.getDirectory(temporary=1, local=1)
        localFiles = self.globDir(localTestDir, sourcePattern)
        if not localFiles:
            logDir = test.getDirectory(temporary=1)
            if logDir != localTestDir:
                logDirFiles = self.globDir(logDir, sourcePattern)
                if logDirFiles:
                    return logDir, logDirFiles
        return localTestDir, localFiles

    def globDir(self, testDir, sourcePattern):
        origCwd = os.getcwd()
        os.chdir(testDir)
        result = glob.glob(sourcePattern)
        os.chdir(origCwd)
        return [os.path.join(testDir, f) for f in result]

    def findPaths(self, test, sourcePattern):
        self.diag.info("Looking for pattern " + sourcePattern + " for " + repr(test))
        testDir, paths = self.glob(test, sourcePattern)
        paths.sort()
        existingPaths = list(filter(os.path.isfile, paths))
        if sourcePattern == "*":  # interpret this specially to mean 'all files which are not collated already'
            return testDir, [f for f in existingPaths if not self.alreadyCollated(test, f, sourcePattern)]
        else:
            return testDir, existingPaths

    def runCollationScript(self, args, test, stdin, stdout, stderr, useShell):
        # Windows isn't clever enough to know how to run Python/Java programs without some help...
        if os.name == "nt" and not useShell:
            interpreter = plugins.getInterpreter(args[0])
            if interpreter:
                args = [interpreter] + args

        try:
            runEnv = test.getRunEnvironment()
            libexecPaths = [os.path.join(p, "libexec") for p in plugins.installationRoots]
            runEnv["PATH"] += os.pathsep + os.pathsep.join(libexecPaths)
            return subprocess.Popen(args, env=runEnv,
                                    stdin=stdin, stdout=stdout, stderr=stderr,
                                    cwd=test.getDirectory(temporary=1), shell=useShell)
        except OSError:
            # Might just be pipe identifiers here
            if hasattr(stdout, "close"):
                stdout.close()
                stderr.close()

    def kill(self, test, sig):
        if self.collationProc:
            proc = self.collationProc
            self.collationProc = None
            killSubProcessAndChildren(proc, cmd=test.getConfigValue("kill_command"))

    def extract(self, test, sourceFiles, targetFile, collationErrFile):
        stem = os.path.splitext(os.path.basename(targetFile))[0]
        scripts = test.getCompositeConfigValue("collate_script", stem)
        sourceFilesStr = ",".join(sourceFiles)
        if len(scripts) == 0:
            if len(sourceFiles) > 1:
                msg = "Multiple files are found for '" + stem + "' in " + \
                    repr(test) + ", but no collate_script is defined.\n"
                sys.stderr.write(msg)
            return shutil.copyfile(sourceFiles[0], targetFile)

        self.collationProc = None
        stdin = None
        for script in scripts:
            args = script.split()
            if self.collationProc:
                stdin = self.collationProc.stdout
            else:
                args += sourceFiles
            self.diag.info("Opening extract process with args " + repr(args))
            if script is scripts[-1]:
                stdout = open(targetFile, "w")
                stderr = open(collationErrFile, "w")
            else:
                stdout = subprocess.PIPE
                stderr = subprocess.STDOUT

            useShell = os.name == "nt" and len(scripts) == 1
            self.collationProc = self.runCollationScript(args, test, stdin, stdout, stderr, useShell)
            if not self.collationProc:
                if os.path.isfile(targetFile):
                    os.remove(targetFile)
                errorMsg = "Could not find extract script '" + script + \
                    "', not extracting file(s) at\n" + sourceFilesStr + "\n"
                stderr = open(collationErrFile, "w")
                stderr.write(errorMsg)
                plugins.printWarning(errorMsg.strip())
                stderr.close()
                return

        if self.collationProc:
            self.diag.info("Waiting for collation process to terminate...")
            self.collationProc.wait()
            if self.collationProc:
                self.collationProc = None
            else:
                procName = args[0]
                briefText = "KILLED (" + os.path.basename(procName) + ")"
                freeText = "Killed collation script '" + procName + \
                    "'\n while collating file(s) at " + sourceFilesStr + "\n"
                test.changeState(Killed(briefText, freeText, test.state))
            stdout.close()
            stderr.close()

        if len(sourceFiles) > 0 and any((os.path.getsize(fn) > 0 for fn in sourceFiles)) and os.path.getsize(targetFile) == 0 and os.path.getsize(collationErrFile) == 0:
            # Collation scripts that don't write anything shouldn't produce empty files...
            # If they write errors though, we might want to pick those up
            os.remove(targetFile)

        collateErrMsg = test.app.filterErrorText(collationErrFile)
        if collateErrMsg:
            msg = "Errors occurred running collate_script(s) " + " and ".join(scripts) + \
                  "\nwhile trying to extract file(s) at \n" + sourceFilesStr + " : \n" + collateErrMsg
            plugins.printWarning(msg)


class FindExecutionHosts(plugins.Action):
    def __call__(self, test):
        test.state.executionHosts = self.getExecutionMachines(test)

    def getExecutionMachines(self, test):
        runMachine = test.app.getRunMachine()
        if runMachine == "localhost":
            return [plugins.gethostname()]
        else:
            return [runMachine]


class CreateCatalogue(plugins.Action):
    def __init__(self):
        self.catalogues = {}
        self.diag = logging.getLogger("catalogues")

    def __call__(self, test):
        if test.getConfigValue("create_catalogues") != "true":
            return

        if test in self.catalogues:
            self.diag.info("Creating catalogue change file...")
            self.createCatalogueChangeFile(test)
        else:
            self.diag.info("Collecting original information...")
            self.catalogues[test] = self.findAllPaths(test)[0]

    def createCatalogueChangeFile(self, test):
        oldPaths = self.catalogues[test]
        newPaths, ignoredPaths = self.findAllPaths(test)
        tmpDir = test.getDirectory(temporary=1, local=1)
        pathsLost, pathsEdited, pathsGained = self.findDifferences(oldPaths, newPaths, ignoredPaths, tmpDir)
        processesGained = self.findProcessesGained(test)
        fileName = test.makeTmpFileName("catalogue")
        with open(fileName, "w") as file:
            if len(pathsLost) == 0 and len(pathsEdited) == 0 and len(pathsGained) == 0:
                file.write("No files or directories were created, edited or deleted.\n")
            if len(pathsGained) > 0:
                file.write("The following new files/directories were created:\n")
                self.writeFileStructure(file, pathsGained)
            if len(pathsEdited) > 0:
                file.write("\nThe following existing files/directories changed their contents:\n")
                self.writeFileStructure(file, pathsEdited)
            if len(pathsLost) > 0:
                file.write("\nThe following existing files/directories were deleted:\n")
                self.writeFileStructure(file, pathsLost)
            if len(processesGained) > 0:
                file.write("\nThe following processes were created:\n")
                self.writeProcesses(file, processesGained)

    def writeProcesses(self, file, processesGained):
        for process in processesGained:
            file.write("- " + process + "\n")

    def writeFileStructure(self, file, pathNames):
        prevParts = []
        tabSize = 4
        for pathName in pathNames:
            parts = pathName.split(os.sep)
            indent = 0
            for index in range(len(parts)):
                part = parts[index]
                indent += tabSize
                if index >= len(prevParts) or part != prevParts[index]:
                    prevParts = []
                    file.write(part + "\n")
                    if index != len(parts) - 1:
                        file.write(("-" * indent))
                else:
                    file.write("-" * tabSize)
            prevParts = parts

    def findProcessesGained(self, test):
        searchString = test.getConfigValue("catalogue_process_string")
        if len(searchString) == 0:
            return []
        # Code untested and unlikely to work on Windows...
        processes = []
        logFile = test.makeTmpFileName(test.getConfigValue("log_file"))
        if not os.path.isfile(logFile):
            return []
        for line in open(logFile):
            if line.startswith(searchString):
                parts = line.strip().split(" : ")
                try:
                    processId = int(parts[-1])
                except ValueError:
                    continue
                self.diag.info("Found process ID " + str(processId))
                if killProcessAndChildren(processId):
                    processes.append(parts[1])
        return processes

    def findAllPaths(self, test):
        allPaths = OrderedDict()
        paths, ignoredPaths = test.listUnownedTmpPaths()
        for path in paths:
            editInfo = self.getEditInfo(path)
            self.diag.info("Path " + path + " edit info " + editInfo)
            allPaths[path] = editInfo
        return allPaths, ignoredPaths

    def getEditInfo(self, fullPath):
        # Check modified times for files and directories, targets for links
        if os.path.islink(fullPath):
            return os.path.realpath(fullPath)
        else:
            return time.strftime(plugins.datetimeFormat, time.localtime(plugins.modifiedTime(fullPath)))

    def findDifferences(self, oldPaths, newPaths, ignoredPaths, writeDir):
        pathsGained, pathsEdited, pathsLost = [], [], []
        for path, modTime in list(newPaths.items()):
            if path not in oldPaths:
                pathsGained.append(self.outputPathName(path, writeDir))
            elif oldPaths[path] != modTime:
                pathsEdited.append(self.outputPathName(path, writeDir))
        for path, modTime in list(oldPaths.items()):
            if path not in newPaths:
                pathsLost.append(self.outputPathName(path, writeDir))
        # Clear out duplicates
        ignoredOutputPaths = [self.outputPathName(path, writeDir) for path in ignoredPaths]
        self.removeParents(pathsEdited, pathsGained)
        self.removeParents(pathsEdited, pathsEdited)
        self.removeParents(pathsEdited, pathsLost)
        self.removeParents(pathsEdited, ignoredOutputPaths)
        self.removeParents(pathsGained, pathsGained)
        self.removeParents(pathsLost, pathsLost)
        return pathsLost, pathsEdited, pathsGained

    def removeParents(self, toRemove, toFind):
        removeList = []
        for path in toFind:
            parent = os.path.split(path)[0]
            if parent in toRemove and not parent in removeList:
                removeList.append(parent)
        for path in removeList:
            self.diag.info("Removing parent path " + path)
            toRemove.remove(path)

    def outputPathName(self, path, writeDir):
        self.diag.info("Output name for " + path)
        return path.replace(writeDir, "<Test Directory>")


class MachineInfoFinder:
    def allMachinesTestPerformance(self, test, fileStem):
        perfMachines = self.findPerformanceMachines(test, fileStem)
        return "any" in perfMachines or all((host in perfMachines for host in test.state.executionHosts))

    def findPerformanceMachines(self, test, fileStem):
        return test.getCompositeConfigValue("performance_test_machine", fileStem)

    def setUpApplication(self, app):
        pass

    def getMachineInformation(self, testArg):
        # A space for subclasses to write whatever they think is relevant about
        # the machine environment right now.
        return ""


class PerformanceFileCreator(plugins.Action):
    def __init__(self, machineInfoFinder):
        self.diag = logging.getLogger("makeperformance")
        self.machineInfoFinder = machineInfoFinder

    def setUpApplication(self, app):
        self.machineInfoFinder.setUpApplication(app)

    def allMachinesTestPerformance(self, test, fileStem):
        return self.machineInfoFinder.allMachinesTestPerformance(test, fileStem)

    def __call__(self, test):
        return self.makePerformanceFiles(test)


class UNIXPerformanceInfoFinder:
    def __init__(self, diag):
        self.diag = diag
        self.includeSystemTime = 0

    def findTimesUsedBy(self, test):
        # Read the UNIX performance file, allowing us to discount system time.
        tmpFile = test.makeTmpFileName("unixperf", forFramework=1)
        self.diag.info("Reading performance file " + tmpFile)
        if not os.path.isfile(tmpFile):
            return None, None

        file = open(tmpFile)
        cpuTime = None
        realTime = None
        for line in file.readlines():
            self.diag.info("Parsing line " + line.strip())
            if line.startswith("user") or line.startswith("User"):
                cpuTime = self.parseUnixTime(line)
            if self.includeSystemTime and (line.startswith("sys") or line.startswith("Sys")):
                cpuTime = cpuTime + self.parseUnixTime(line)
            if line.startswith("real") or line.startswith("Real"):
                realTime = self.parseUnixTime(line)
        return cpuTime, realTime

    def parseUnixTime(self, line):
        # Assumes output of GNU time
        words = line.strip().split()
        return float(words[-1])

    def setUpApplication(self, app):
        self.includeSystemTime = app.getConfigValue("cputime_include_system_time")

# Class for making a performance file directly from system-collected information,
# rather than parsing reported entries in a log file


class MakePerformanceFile(PerformanceFileCreator):
    def __init__(self, machineInfoFinder):
        PerformanceFileCreator.__init__(self, machineInfoFinder)
        self.systemPerfInfoFinder = UNIXPerformanceInfoFinder(self.diag)

    def setUpApplication(self, app):
        PerformanceFileCreator.setUpApplication(self, app)
        self.systemPerfInfoFinder.setUpApplication(app)

    def makePerformanceFiles(self, test):
        cpuTime, realTime = self.systemPerfInfoFinder.findTimesUsedBy(test)
        # There was still an error (jobs killed in emergency), so don't write performance files
        if cpuTime is None:
            return

        fileToWrite = test.makeTmpFileName("performance")
        self.writeFile(test, cpuTime, realTime, fileToWrite)

    def timeString(self, timeVal):
        return str(round(float(timeVal), 1)).rjust(9)

    def writeFile(self, test, cpuTime, realTime, fileName):
        file = open(fileName, "w")
        cpuLine = "CPU time   : " + self.timeString(cpuTime) + " sec. " + test.state.hostString() + "\n"
        file.write(cpuLine)
        if realTime is not None:
            realLine = "Real time  : " + self.timeString(realTime) + " sec.\n"
            file.write(realLine)
        file.write(self.machineInfoFinder.getMachineInformation(test))

# Relies on the config entry performance_logfile_extractor, so looks in the log file for anything reported
# by the program


class ExtractPerformanceFiles(PerformanceFileCreator):
    def __init__(self, machineInfoFinder):
        PerformanceFileCreator.__init__(self, machineInfoFinder)
        self.entryFiles = None
        self.logFileStem = None

    def makePerformanceFiles(self, test):
        entryFinders = test.getConfigValue("performance_logfile_extractor")
        entryFiles = test.getConfigValue("performance_logfile")
        defaultLogFileStem = test.getConfigValue("log_file")
        self.diag.info("Found the following entry finders:" + str(entryFinders))
        for fileStem, entryFinder in list(entryFinders.items()):
            if len(entryFinder) == 0:
                continue  # don't allow empty entry finders
            if not self.allMachinesTestPerformance(test, fileStem):
                self.diag.info("Not extracting performance file for " + fileStem + ": not on performance machines")
                continue
            values = []
            logFileStems = entryFiles.get(fileStem, [defaultLogFileStem])
            for logFileStem in logFileStems:
                self.diag.info("Looking for log files matching " + logFileStem)
                for fileName in self.findLogFiles(test, logFileStem):
                    self.diag.info("Scanning log file for entry: " + entryFinder)
                    values += self.findValues(fileName, entryFinder)
            if len(values) > 0:
                fileName = self.getFileToWrite(test, fileStem)
                self.diag.info("Writing performance to file " + fileName)
                contents = self.makeFileContents(test, values, fileStem)
                self.saveFile(fileName, contents)

    def getFileToWrite(self, test, stem):
        return test.makeTmpFileName(stem)

    def findLogFiles(self, test, stem):
        collatedfiles = glob.glob(test.makeTmpFileName(stem))
        if len(collatedfiles) == 0:
            return glob.glob(test.makeTmpFileName(stem, forComparison=0))
        return collatedfiles

    def saveFile(self, fileName, contents):
        file = open(fileName, "w")
        file.write(contents)
        file.close()

    def makeFileContents(self, test, values, fileStem):
        # Round to accuracy 0.01
        unit = test.getCompositeConfigValue("performance_unit", fileStem)
        if fileStem.find("mem") != -1:
            return self.makeMemoryLine(values, fileStem) + " " + unit + "\n"
        else:
            return self.makeTimeLine(values, fileStem) + " " + unit + self.getMachineContents(test)

    def getMachineContents(self, test):
        return " " + test.state.hostString() + "\n" + self.machineInfoFinder.getMachineInformation(test)

    def makeMemoryLine(self, values, fileStem):
        maxVal = max(values)
        roundedMaxVal = float(int(100*maxVal))/100
        return "Max " + fileStem.capitalize() + "  :      " + str(roundedMaxVal)

    def makeTimeLine(self, values, fileStem):
        sum = 0.0
        for value in values:
            sum += value
        roundedSum = float(int(10*sum))/10
        return "Total " + fileStem.capitalize() + "  :      " + str(roundedSum)

    def findValues(self, logFile, entryFinder):
        values = []
        for line in open(logFile):
            value = self.getValue(line, entryFinder)
            if value is not None:
                self.diag.info(" found value: " + str(value))
                values.append(value)
        return values

    def getValue(self, line, entryFinder):
        match = re.search(entryFinder, line)
        if match is not None:
            # If there are groups in the string, go through them till we find one we can parse
            for group in match.groups():
                groupValue = self.parseString(group)
                if groupValue is not None:
                    return groupValue
            # locates the first whitespace after an occurrence of entryFinder in line,
            # and scans the rest of the string after that whitespace
            return self.getValue(line, entryFinder + r'\S*\s+(\S+)')

    def parseString(self, stringValue):
        try:
            return float(stringValue)
        except ValueError:
            return self.parseAsTimeString(stringValue)

    def parseAsTimeString(self, stringValue):
        # try parsing the memString as a h*:mm:ss time string
        # * - any number of figures are allowed for the hour part
        timeRegExp = re.compile(r'(?P<hours>\d+)\:(?P<minutes>\d\d)\:(?P<seconds>\d\d)')
        match = timeRegExp.match(stringValue)
        if match:
            hours = float(match.group('hours'))
            minutes = float(match.group('minutes'))
            seconds = float(match.group('seconds'))
            return hours*60*60 + minutes*60 + seconds



#!/usr/bin/env python

# texttest_release.py : makes a release from Bazaar source control. Can do so from a local setup using the
# environment variable BZRROOT (modelled on the layout at Jeppesen), or directly from Launchpad.

# Usage texttest_release.py [ -v <release_name> ] [ -t <tag> ] [ -x ] [ -d <working_dir> ]

# <working_dir> indicates where temporary files are written and the final zip file will end up.
# It defaults to the current working directory.

# <release_name> defaults to "current" and should be overridden when making external releases

# <tag> indicates a version control tag to use. This should be a pre-existing one that is set
# in both the TextTest source and the self-tests (but not StoryText, currently)

# The -x flag should be provided if the temporary files are to be left. Mostly useful for testing.

import os
import sys
import shutil
import subprocess
import stat
import time
from glob import glob
from getopt import getopt


def getSfUser():
    userMap = {"geoff": "gjb1002", "E601429": "gjb1002", "dgalda": "dagal"}
    user = os.getenv("USER", os.getenv("USERNAME"))
    return userMap.get(user, user)


def system(args):
    print(" ".join(args))
    subprocess.call(args)


def getBzrLocation(product, artefact):
    # If $BZRROOT set, use the local repositories as laid out at Jeppesen, otherwise use Launchpad
    bzrRoot = os.getenv("BZRROOT", "/carm/proj/texttest/bzr")
    if bzrRoot and os.path.isdir(bzrRoot):
        return os.path.join(bzrRoot, product, artefact, "branches/HEAD")
    else:
        launchpadNames = {"source": "trunk", "tests": "selftest-trunk"}
        return "lp:~geoff.bache/" + product.lower() + "/" + launchpadNames.get(artefact)


def exportFromBzr(dest, tagName):
    args = ["bzr", "checkout", "--lightweight"]
    if tagName:
        args.append("-rtag:" + tagName)
    args.append(getBzrLocation("TextTest", "source"))
    args.append(dest)
    system(args)


def createSource(reldir, debug):
    versionFile = os.path.join(reldir, "texttestlib", "texttest_version.py")
    updateVersionFile(versionFile, releaseName)
    args = ["python", "setup.py", "sdist"]
    if not debug:
        args.append("upload")
    subprocess.call(args, cwd=reldir)


def updateVersionFile(versionFile, releaseName):
    newFileName = versionFile + ".new"
    with open(newFileName, "w") as newFile:
        for line in open(versionFile):
            newFile.write(line.replace("trunk", releaseName))
    os.remove(versionFile)
    os.rename(newFileName, versionFile)


def getCommandLine():
    options, leftovers = getopt(sys.argv[1:], "d:v:t:x")
    optDict = dict(options)
    return optDict.get("-d", os.getcwd()), optDict.get("-v", "current"), optDict.get("-t", ""), "-x" in optDict


def readmeUpdated(readme):
    if not os.path.isfile(readme):
        return False

    mtime = os.stat(readme)[stat.ST_MTIME]
    currTime = time.time()
    hoursSinceEdit = (currTime - mtime) / 3600
    return hoursSinceEdit < 24


if __name__ == "__main__":
    rootDir, releaseName, tagName, debug = getCommandLine()
    devRelease = "dev" in releaseName
    readme = os.path.join(rootDir, "readme.txt")
    if not devRelease and not readmeUpdated(readme):
        print("Cannot make external release, readme file at", readme, "has not been updated recently.")
        sys.exit(1)
    reldir = "texttest-" + releaseName
    actualRoot = os.path.join(rootDir, reldir)
    if os.path.isdir(actualRoot):
        shutil.rmtree(actualRoot)
    os.makedirs(actualRoot)

    exportFromBzr(actualRoot, tagName)
    createSource(actualRoot, debug)

    if not devRelease:
        tarball = os.path.join(actualRoot, "dist", "TextTest-" + releaseName + ".tar.gz")
        target = os.path.join("/home/frs/project/texttest/texttest", releaseName)
        args = ["rsync", "-av", "--rsh=ssh", tarball, readme, getSfUser() + "@web.sourceforge.net:" + target]
        print(" ".join(args))
        if not debug:
            subprocess.call(args)




""" Base class for all the queue system implementations """

import subprocess, os, sys
from locale import getpreferredencoding
from texttestlib import plugins


class QueueSystem(object):
    def __init__(self, *args):
        pass

    def submitSlaveJob(self, cmdArgs, slaveEnv, logDir, submissionRules, jobType):
        try:
            process = subprocess.Popen(cmdArgs, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                       cwd=logDir, env=self.getSlaveEnvironment(slaveEnv),
                                       encoding=getpreferredencoding(), startupinfo=plugins.getHideStartUpInfo())
            stdout, stderr = process.communicate()
            errorMessage = self.findErrorMessage(stderr, cmdArgs, jobType)
        except OSError:
            errorMessage = self.getFullSubmitError(
                "local machine is not a submit host: running '" + cmdArgs[0] + "' failed.", cmdArgs, jobType)
        if errorMessage:
            return None, errorMessage
        else:
            return self.findJobId(stdout), None

    def supportsPolling(self):
        return True

    def findErrorMessage(self, stderr, *args):
        if len(stderr) > 0:
            basicError = self.findSubmitError(stderr)
            if basicError:
                return self.getFullSubmitError(basicError, *args)

    def getFullSubmitError(self, errorMessage, cmdArgs, jobType):
        qname = self.getQueueSystemName()
        err = "Failed to submit "
        if jobType:
            err += jobType + " "
        err += "to " + qname + " (" + errorMessage.strip() + ")\n" + \
               "Submission command was '" + self.formatCommand(cmdArgs) + "'\n"
        return err

    def getSlaveEnvironment(self, env):
        if len(env) >= len(os.environ):  # full environment sent
            return env
        else:
            return self.makeSlaveEnvironment(env)

    def getSlaveVarsToBlock(self):
        return []

    def getCapacity(self):
        pass  # treated as no restriction

    def setRemoteProcessId(self, *args):
        pass  # only cloud cares about this

    def getRemoteTestMachine(self, *args):
        pass  # only cloud cares about this

    def cleanup(self, *args):
        return True  # only cloud cares about this

    def slavesOnRemoteSystem(self):
        return False

    def getWindowsExecutable(self):
        # sys.executable could be something other than Python... like storytext. Don't involve that here
        if os.path.basename(sys.executable) in ["python.exe", "python2.exe", "python3.exe"]:
            return sys.executable
        else:
            path = os.path.join(sys.exec_prefix, "Scripts", "python.exe")
            if os.path.isfile(path):
                return path
            else:
                return os.path.join(sys.exec_prefix, "python.exe")

    def getTextTestArgs(self):
        texttest = plugins.getTextTestProgram()
        if getattr(sys, 'frozen', False):
            return [texttest.replace("texttest.exe", "texttestc.exe")]
        return [self.getWindowsExecutable(), texttest] if os.name == "nt" else [texttest]

    def makeSlaveEnvironment(self, env):
        newEnv = plugins.copyEnvironment(ignoreVars=self.getSlaveVarsToBlock())
        for var, value in list(env.items()):
            newEnv[var] = value
        return newEnv

    def getQueueSystemName(self):
        modname = self.__class__.__module__
        return modname.split(".")[-1].upper()

    def addExtraAndCommand(self, args, submissionRules, commandArgs):
        args += submissionRules.getExtraSubmitArgs()
        if commandArgs:
            args += self.shellWrapArgs(commandArgs)
        return args

    def formatCommand(self, cmdArgs):
        return " ".join(cmdArgs[:-2]) + " ... "

    def getSubmitCmdArgs(self, submissionRules, commandArgs=[], slaveEnv={}):
        return commandArgs

    def makeHeader(self, text):
        return "-" * 10 + " " + text + " " + "-" * 10 + "\n"

    def getJobFailureInfo(self, jobId):
        name = self.getQueueSystemName()
        header = self.makeHeader("Full accounting info from " + name)
        if jobId is None:
            return header + "No job has been submitted to " + name
        else:
            return header + self._getJobFailureInfo(jobId)

    def shellWrapArgs(self, commandArgs):
        # Must use exec so as not to create extra processes: SGE's qdel isn't very clever when
        # it comes to noticing extra shells
        return ["exec", "$SHELL", "-c", "exec {}".format(plugins.commandLineString(commandArgs, defaultQuoteChar="'"))]




"""
Code associated with the left-hand tree view for tests 
"""
from gi.repository import Gtk, GObject, Pango
from . import guiutils
import logging
from texttestlib import plugins
from collections import OrderedDict


class TestColumnGUI(guiutils.SubGUI):
    def __init__(self, dynamic, testCount):
        guiutils.SubGUI.__init__(self)
        self.addedCount = 0
        self.totalNofTests = testCount
        self.totalNofDistinctTests = testCount
        self.nofSelectedTests = 0
        self.nofDistinctSelectedTests = 0
        self.totalNofTestsShown = 0
        self.versionString = ""
        self.column = None
        self.dynamic = dynamic
        self.testSuiteSelection = False
        self.diag = logging.getLogger("Test Column GUI")
        self.allSuites = []

    def addSuites(self, suites):
        self.allSuites = suites

    def createView(self):
        testRenderer = Gtk.CellRendererText()
        self.column = Gtk.TreeViewColumn(self.getTitle(), testRenderer, text=0, background=1, foreground=7)
        self.column.name = "Test Name"  # Not a widget, so we can't set a name, do this instead
        self.column.set_resizable(True)
        self.column.set_cell_data_func(testRenderer, self.renderSuitesBold)
        if not self.dynamic:
            self.column.set_clickable(True)
            self.column.connect("clicked", self.columnClicked)
        if guiutils.guiConfig.getValue("auto_sort_test_suites") == 1:
            self.column.set_sort_indicator(True)
            self.column.set_sort_order(Gtk.SortType.ASCENDING)
        elif guiutils.guiConfig.getValue("auto_sort_test_suites") == -1:
            self.column.set_sort_indicator(True)
            self.column.set_sort_order(Gtk.SortType.DESCENDING)
        return self.column

    def renderSuitesBold(self, dummy, cell, model, iter, data):
        if model.get_value(iter, 2)[0].classId() == "test-case":
            cell.set_property('font', "")
        else:
            cell.set_property('font', "bold")

    def columnClicked(self, *args):
        if not self.column.get_sort_indicator():
            self.column.set_sort_indicator(True)
            self.column.set_sort_order(Gtk.SortType.ASCENDING)
            order = 1
        else:
            order = self.column.get_sort_order()
            if order == Gtk.SortType.ASCENDING:
                self.column.set_sort_order(Gtk.SortType.DESCENDING)
                order = -1
            else:
                self.column.set_sort_indicator(False)
                order = 0

        self.notify("ActionStart")
        self.setSortingOrder(order)
        if order == 1:
            self.notify("Status", "Tests sorted in alphabetical order.")
        elif order == -1:
            self.notify("Status", "Tests sorted in descending alphabetical order.")
        else:
            self.notify("Status", "Tests sorted according to testsuite file.")
        self.notify("RefreshTestSelection")
        self.notify("ActionStop")

    def setSortingOrder(self, order, suite=None):
        if not suite:
            for suite in self.allSuites:
                self.setSortingOrder(order, suite)
        else:
            self.notify("Status", "Sorting suite " + suite.name + " ...")
            self.notify("ActionProgress")
            suite.autoSortOrder = order
            suite.updateOrder()
            for test in suite.testcases:
                if test.classId() == "test-suite":
                    self.setSortingOrder(order, test)

    def getTitle(self):
        title = "Tests: "
        if self.versionString and len(self.versionString) > 40:
            reducedVersionString = self.versionString[:40] + "..."
        else:
            reducedVersionString = self.versionString

        if self.testSuiteSelection:
            # We don't care about totals with test suites
            title += plugins.pluralise(self.nofSelectedTests, "suite") + " selected"
            if self.versionString:
                title += ", " + reducedVersionString
            elif self.nofDistinctSelectedTests != self.nofSelectedTests:
                title += ", " + str(self.nofDistinctSelectedTests) + " distinct"
            return title

        if self.nofSelectedTests == self.totalNofTests:
            title += "All " + str(self.totalNofTests) + " selected"
        else:
            title += str(self.nofSelectedTests) + "/" + str(self.totalNofTests) + " selected"

        if not self.dynamic:
            if self.versionString:
                title += ", " + reducedVersionString
            elif self.totalNofDistinctTests != self.totalNofTests:
                if self.nofDistinctSelectedTests == self.totalNofDistinctTests:
                    title += ", all " + str(self.totalNofDistinctTests) + " distinct"
                else:
                    title += ", " + str(self.nofDistinctSelectedTests) + "/" + \
                        str(self.totalNofDistinctTests) + " distinct"

        if self.totalNofTestsShown == self.totalNofTests:
            if self.dynamic and self.totalNofTests > 0:
                title += ", none hidden"
        elif self.totalNofTestsShown == 0:
            title += ", all hidden"
        else:
            title += ", " + str(self.totalNofTests - self.totalNofTestsShown) + " hidden"

        return title

    def updateTitle(self, initial=False):
        if self.column:
            self.column.set_title(self.getTitle())

    def notifyTestTreeCounters(self, totalDelta, totalShownDelta, totalRowsDelta, initial=False):
        self.addedCount += totalDelta
        if not initial or self.totalNofTests < self.addedCount:
            self.totalNofTests += totalDelta
            self.totalNofDistinctTests += totalRowsDelta
        self.totalNofTestsShown += totalShownDelta
        self.updateTitle(initial)

    def notifyAllRead(self):
        if self.addedCount != self.totalNofTests:
            self.totalNofTests = self.addedCount
            self.updateTitle()

    def countTests(self, tests):
        if self.dynamic:
            return len(tests), False

        testCount, suiteCount = 0, 0
        for test in tests:
            if test.classId() == "test-case":
                testCount += 1
            else:
                suiteCount += 1
        if suiteCount and not testCount:
            return suiteCount, True
        else:
            return testCount, False

    def getVersionString(self, tests, distinctTestCount):
        if not self.dynamic and distinctTestCount == 1 and self.totalNofTests != self.totalNofDistinctTests:
            versions = [test.app.getFullVersion().replace("_", "__") or "<default>" for test in tests]
            return "version" + ("s" if len(versions) > 1 else "") + " " + ",".join(versions)
        else:
            return ""

    def notifyNewTestSelection(self, tests, dummyApps, distinctTestCount, *args, **kw):
        if self.dynamic:
            tests = [t for t in tests if t.classId() == "test-case"]
        self.updateTestInfo(tests, distinctTestCount)

    def updateTestInfo(self, tests, distinctTestCount):
        newCount, suitesOnly = self.countTests(tests)
        if distinctTestCount > newCount:
            distinctTestCount = newCount
        newVersionStr = self.getVersionString(tests, distinctTestCount)
        if self.nofSelectedTests != newCount or newVersionStr != self.versionString or \
                self.nofDistinctSelectedTests != distinctTestCount or suitesOnly != self.testSuiteSelection:
            self.diag.info("New selection count = " + repr(newCount) + ", distinct = " +
                           str(distinctTestCount) + ", test suites only = " + repr(suitesOnly))
            self.nofSelectedTests = newCount
            self.nofDistinctSelectedTests = distinctTestCount
            self.testSuiteSelection = suitesOnly
            self.versionString = newVersionStr
            self.updateTitle()

    def notifyVisibility(self, tests, newValue):
        testCount = sum((int(test.classId() == "test-case") for test in tests))
        if newValue:
            self.totalNofTestsShown += testCount
        else:
            self.totalNofTestsShown -= testCount
        self.updateTitle()


class TestIteratorMap:
    def __init__(self, dynamic, allApps):
        self.dict = OrderedDict()
        self.dynamic = dynamic
        self.parentApps = {}
        for app in allApps:
            for extra in [app] + app.extras:
                self.parentApps[extra] = app

    def getKey(self, test):
        if self.dynamic:
            return test
        elif test is not None:
            return self.parentApps.get(test.app, test.app), test.getRelPath()

    def store(self, test, iter):
        self.dict[self.getKey(test)] = iter

    def updateIterator(self, test, oldRelPath):
        # relative path of test has changed
        key = self.parentApps.get(test.app, test.app), oldRelPath
        iter = self.dict.get(key)
        if iter is not None:
            self.store(test, iter)
            del self.dict[key]
            return iter
        else:
            return self.getIterator(test)

    def getIterator(self, test):
        return self.dict.get(self.getKey(test))

    def remove(self, test):
        key = self.getKey(test)
        if key in self.dict:
            del self.dict[key]


class TestTreeGUI(guiutils.ContainerGUI):
    def __init__(self, dynamic, allApps, popupGUI, subGUI):
        guiutils.ContainerGUI.__init__(self, [subGUI])
        self.model = Gtk.TreeStore(GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_PYOBJECT,
                                   GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_BOOLEAN,
                                   GObject.TYPE_STRING, GObject.TYPE_STRING)
        self.popupGUI = popupGUI
        self.itermap = TestIteratorMap(dynamic, allApps)
        self.selection = None
        self.selecting = False
        self.selectedTests = []
        self.clipboardTests = set()
        self.dynamic = dynamic
        self.collapseStatic = 100 if dynamic else guiutils.guiConfig.getValue("static_collapse_suites")
        self.filteredModel = None
        self.treeView = None
        self.newTestsVisible = guiutils.guiConfig.showCategoryByDefault("not_started")
        self.diag = logging.getLogger("Test Tree")
        self.longActionRunning = False
        self.recreateOnActionStop = False
        self.testSuitesWithResults = set()

    def notifyDefaultVisibility(self, newValue):
        self.newTestsVisible = newValue

    def isExpanded(self, iter):
        parentIter = self.filteredModel.iter_parent(iter)
        return not parentIter or self.treeView.row_expanded(self.filteredModel.get_path(parentIter))

    def notifyAllRead(self, *args):
        if not self.dynamic:
            self.newTestsVisible = True
            self.model.foreach(self.makeRowVisible)
            if self.collapseStatic != -1:
                self.expandLevel(self.treeView, self.filteredModel.get_iter_first(), self.collapseStatic)
            else:
                self.treeView.expand_all()
        self.notify("AllRead")

    def makeRowVisible(self, model, dummyPath, iter):
        model.set_value(iter, 5, True)

    def getNodeName(self, suite, parent):
        nodeName = suite.name
        if parent == None:
            appName = suite.app.name + suite.app.versionSuffix()
            if appName != nodeName:
                nodeName += " (" + appName + ")"
        return nodeName

    def addSuiteWithParent(self, suite, parent, follower=None):
        nodeName = self.getNodeName(suite, parent)
        self.diag.info("Adding node with name " + nodeName)
        colour = guiutils.guiConfig.getTestColour("not_started")
        row = [nodeName, colour, [suite], "", colour, self.newTestsVisible, "", "black"]
        iter = self.model.insert_before(parent, follower, row)
        storeIter = iter.copy()
        self.itermap.store(suite, storeIter)
        path = self.model.get_path(iter)
        if self.newTestsVisible and parent is not None:
            filterPath = self.filteredModel.convert_child_path_to_path(path)
            self.treeView.expand_to_path(filterPath)
        return iter

    def createView(self):
        self.filteredModel = self.model.filter_new()
        self.filteredModel.set_visible_column(5)
        self.treeView = Gtk.TreeView(self.filteredModel)
        self.treeView.set_search_column(0)
        self.treeView.set_name("Test Tree")
        self.treeView.expand_all()

        self.selection = self.treeView.get_selection()
        self.selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        if self.dynamic:
            self.selection.set_select_function(self.canSelect, self)

        testsColumn = self.subguis[0].createView()
        self.treeView.append_column(testsColumn)
        if self.dynamic:
            detailsRenderer = Gtk.CellRendererText()
            detailsRenderer.set_property('wrap-width', 350)
            detailsRenderer.set_property('wrap-mode', Pango.WrapMode.WORD_CHAR)
            recalcRenderer = Gtk.CellRendererPixbuf()
            detailsColumn = Gtk.TreeViewColumn("Details")
            detailsColumn.pack_start(detailsRenderer, True)
            detailsColumn.pack_start(recalcRenderer, False)
            detailsColumn.add_attribute(detailsRenderer, 'text', 3)
            detailsColumn.add_attribute(detailsRenderer, 'background', 4)
            detailsColumn.add_attribute(recalcRenderer, 'stock_id', 6)
            detailsColumn.set_resizable(True)
            guiutils.addRefreshTips(self.treeView, "test", recalcRenderer, detailsColumn, 6)
            self.treeView.append_column(detailsColumn)

        self.treeView.connect('row-expanded', self.rowExpanded)
        self.expandLevel(self.treeView, self.filteredModel.get_iter_first())
        self.treeView.connect("button_press_event", self.popupGUI.showMenu)
        self.selection.connect("changed", self.userChangedSelection)

        self.treeView.show()

        self.popupGUI.createView()
        return self.addScrollBars(self.treeView, hpolicy=Gtk.PolicyType.NEVER)

    def notifyTopWindow(self, *args):
        # avoid the quit button getting initial focus, give it to the tree view (why not?)
        self.treeView.grab_focus()

    @staticmethod
    def canSelect(selection, model, path, is_selected, user_data):
        pathIter = user_data.filteredModel.get_iter(path)
        test = user_data.filteredModel.get_value(pathIter, 2)[0]
        return test.classId() == "test-case" or test in user_data.testSuitesWithResults

    def rowExpanded(self, treeview, iter, path):
        self.expandLevel(treeview, self.filteredModel.iter_children(iter), self.collapseStatic - len(path))

    def userChangedSelection(self, *args):
        if not self.selecting and not hasattr(self.selection, "unseen_changes"):
            self.selectionChanged(direct=True)

    def selectionChanged(self, direct):
        newSelection = self.getSelected()
        if newSelection != self.selectedTests:
            self.sendSelectionNotification(newSelection, direct)
            if direct:
                self.updateRecalculationMarkers()

    def notifyRefreshTestSelection(self):
        # The selection hasn't changed, but we want to e.g.
        # recalculate the action sensitiveness and make sure we can still see the selected tests.
        self.sendSelectionNotification(self.selectedTests)
        self.scrollToFirstTest()

    def notifyRecomputed(self, test):
        iter = self.itermap.getIterator(test)
        # If we've recomputed, clear the recalculation icons
        self.setNewRecalculationStatus(iter, test, [])

    def getSortedSelectedTests(self, suite):
        appTests = [test for test in self.selectedTests if test.app is suite.app]
        allTests = suite.allTestsAndSuites()
        appTests.sort(key=allTests.index)
        return appTests

    def notifyWriteTestsIfSelected(self, suite, file):
        for test in self.getSortedSelectedTests(suite):
            self.writeSelectedTest(test, file)

    def shouldListSubTests(self, test):
        if test.parent is None or not all((self.isVisible(test) for test in test.testCaseList())):
            return True

        filters = test.app.getFilterList([test])
        return len(filters) > 0

    def writeSelectedTest(self, test, file):
        if test.classId() == "test-suite":
            if self.shouldListSubTests(test):
                for subTest in test.testcases:
                    if self.isVisible(subTest):
                        self.writeSelectedTest(subTest, file)
                return
        file.write(test.getRelPath() + "\n")

    def updateRecalculationMarker(self, model, dummy, iter):
        tests = model.get_value(iter, 2)
        if not tests[0].stateInGui.isComplete():
            return

        recalcComparisons = tests[0].stateInGui.getComparisonsForRecalculation()
        childIter = self.filteredModel.convert_iter_to_child_iter(iter)
        self.setNewRecalculationStatus(childIter, tests[0], recalcComparisons)

    def setNewRecalculationStatus(self, iter, test, recalcComparisons):
        oldVal = self.model.get_value(iter, 6)
        newVal = self.getRecalculationIcon(recalcComparisons)
        if newVal != oldVal:
            self.model.set_value(iter, 6, newVal)
        self.notify("Recalculation", test, recalcComparisons, newVal)

    def getRecalculationIcon(self, recalc):
        if recalc:
            return "gtk-refresh"
        else:
            return ""

    def checkRelatedForRecalculation(self, test):
        self.filteredModel.foreach(self.checkRecalculationIfMatches, test)

    def checkRecalculationIfMatches(self, model, path, iter, test):
        tests = model.get_value(iter, 2)
        if tests[0] is not test and tests[0].getRelPath() == test.getRelPath():
            self.updateRecalculationMarker(model, path, iter)

    def getSelectedApps(self, tests):
        apps = []
        for test in tests:
            if test.app not in apps:
                apps.append(test.app)
        return apps

    def notifyActionStart(self, foreground=True):
        if not foreground:
            self.longActionRunning = True

    def notifyActionStop(self, foreground=True):
        if not foreground:
            if self.longActionRunning and self.recreateOnActionStop:
                self.sendActualSelectionNotification(direct=False)
            self.longActionRunning = False
            self.recreateOnActionStop = False

    def sendActualSelectionNotification(self, direct):
        apps = self.getSelectedApps(self.selectedTests)
        self.notify("NewTestSelection", self.selectedTests, apps, self.selection.count_selected_rows(), direct)

    def sendSelectionNotification(self, tests, direct=True):
        if len(tests) < 10:
            self.diag.info("Selection now changed to " + repr(tests))
        else:
            self.diag.info("Selection now of size " + str(len(tests)))
        self.selectedTests = tests
        if self.longActionRunning:
            self.recreateOnActionStop = True
            self.subguis[0].updateTestInfo(tests, self.selection.count_selected_rows())
        else:
            self.sendActualSelectionNotification(direct)

    def updateRecalculationMarkers(self):
        if self.dynamic:
            self.selection.selected_foreach(self.updateRecalculationMarker)

    def getSelected(self):
        allSelected = []
        prevSelected = set(self.selectedTests)

        def addSelTest(model, dummy, iter, selected):
            selected += self.getNewSelected(model.get_value(iter, 2), prevSelected)

        self.selection.selected_foreach(addSelTest, allSelected)
        return allSelected

    def getNewSelected(self, tests, prevSelected):
        intersection = prevSelected.intersection(set(tests))
        if len(intersection) == 0 or len(intersection) == len(tests) or len(intersection) == len(prevSelected):
            return tests
        else:
            return list(intersection)

    def findIter(self, test):
        try:
            childIter = self.itermap.getIterator(test)
            if childIter:
                return self.filteredModel.convert_child_iter_to_iter(childIter)
        except RuntimeError:
            # convert_child_iter_to_iter throws RunTimeError if the row is hidden in the TreeModelFilter
            self.diag.info("Could not find iterator for " + repr(test) + ", possibly row is hidden.")

    def notifySetTestSelection(self, selTests, criteria="", selectCollapsed=True, direct=False):
        actualSelection = self.selectTestRows(selTests, selectCollapsed)
        # Here it's been set via some indirect mechanism, might want to behave differently
        self.sendSelectionNotification(actualSelection, direct=direct)
        self.updateRecalculationMarkers()

    def selectTestRows(self, selTests, selectCollapsed=True):
        self.selecting = True  # don't respond to each individual programmatic change here
        self.selection.unselect_all()
        treeView = self.selection.get_tree_view()
        firstPath = None
        actuallySelected = []
        for test in selTests:
            iterValid, iter = self.findIter(test)
            if not iterValid or (not selectCollapsed and not self.isExpanded(iter)):
                continue

            actuallySelected.append(test)
            path = self.filteredModel.get_path(iter)
            if not firstPath:
                firstPath = path
            if selectCollapsed:
                treeView.expand_to_path(path)
            self.selection.select_iter(iter)
        treeView.grab_focus()
        if firstPath is not None and treeView.get_property("visible"):
            self.scrollToPath(firstPath)
        self.selecting = False
        return actuallySelected

    def scrollToFirstTest(self):
        if len(self.selectedTests) > 0:
            test = self.selectedTests[0]
            iterValid, iter = self.findIter(test)
            path = self.filteredModel.get_path(iter)
            self.scrollToPath(path)

    def scrollToPath(self, path):
        treeView = self.selection.get_tree_view()
        cellArea = treeView.get_cell_area(path, treeView.get_columns()[0])
        visibleArea = treeView.get_visible_rect()
        if cellArea.y < 0 or cellArea.y > visibleArea.height:
            treeView.scroll_to_cell(path, use_align=True, row_align=0.1)

    def expandLevel(self, view, iter, recurseLevel=100):
        # Make sure expanding expands everything, better than just one level as default...
        # Avoid using view.expand_row(path, open_all=True), as the open_all flag
        # doesn't seem to send the correct 'row-expanded' signal for all rows ...
        # This way, the signals are generated one at a time and we call back into here.
        model = view.get_model()
        while iter is not None:
            if recurseLevel > 0:
                view.expand_row(model.get_path(iter), open_all=False)
            iter = view.get_model().iter_next(iter)

    def notifyTestAppearance(self, test, detailText, colour1, colour2, approved):
        iter = self.itermap.getIterator(test)
        self.model.set_value(iter, 1, colour1)
        self.model.set_value(iter, 3, detailText)
        self.model.set_value(iter, 4, colour2)
        if test.classId() == "test-suite":  # Happens with Replace Text sometimes
            self.testSuitesWithResults.add(test)
        if approved:
            self.checkRelatedForRecalculation(test)

    def notifyLifecycleChange(self, test, *args):
        if test in self.selectedTests:
            self.notify("LifecycleChange", test, *args)

    def notifyFileChange(self, test, *args):
        if test in self.selectedTests:
            self.notify("FileChange", test, *args)

    def notifyDescriptionChange(self, test, *args):
        if test in self.selectedTests:
            self.notify("DescriptionChange", test, *args)

    def notifyRefreshFilePreviews(self, test, *args):
        if test in self.selectedTests:
            self.notify("RefreshFilePreviews", test, *args)

    def isVisible(self, test):
        iterValid, filteredIter = self.findIter(test)
        if iterValid:
            return True
        else:
            self.diag.info("No iterator found for " + repr(test))
            return False

    def findAllTests(self):
        tests = []
        self.model.foreach(self.appendTest, tests)
        return tests

    def appendTest(self, model, dummy, iter, tests):
        for test in model.get_value(iter, 2):
            if test.classId() == "test-case":
                tests.append(test)

    def getTestForAutoSelect(self):
        allTests = self.findAllTests()
        if len(allTests) == 1:
            test = allTests[0]
            if self.isVisible(test):
                return test

    def notifyAllComplete(self):
        # Window may already have been closed...
        if self.selection.get_tree_view():
            test = self.getTestForAutoSelect()
            if test:
                actualSelection = self.selectTestRows([test])
                self.sendSelectionNotification(actualSelection)

    def notifyAdd(self, test, initial):
        if test.classId() == "test-case":
            self.notify("TestTreeCounters", initial=initial, totalDelta=1,
                        totalShownDelta=self.getTotalShownDelta(), totalRowsDelta=self.getTotalRowsDelta(test))
        elif self.dynamic and test.isEmpty():
            return  # don't show empty suites in the dynamic GUI

        self.diag.info("Adding test " + repr(test))
        self.tryAddTest(test, initial)
        if test.parent is None and not initial:
            # We've added a new suite, we should also select it as it's likely the user wants to add stuff under it
            # Also include the knock-on effects, i.e. selecting the test tab etc
            self.notifySetTestSelection([test], direct=True)

    def notifyClipboard(self, tests, cut=False):
        if cut:
            colourKey = "clipboard_cut"
        else:
            colourKey = "clipboard_copy"
        colour = guiutils.guiConfig.getTestColour(colourKey)
        toRemove = self.clipboardTests.difference(set(tests))
        self.clipboardTests = set(tests)
        for test in tests:
            iter = self.itermap.getIterator(test)
            self.model.set_value(iter, 7, colour)
        for test in toRemove:
            iter = self.itermap.getIterator(test)
            if iter:
                self.model.set_value(iter, 7, "black")

    def getTotalRowsDelta(self, test):
        if self.itermap.getIterator(test):
            return 0
        else:
            return 1

    def getTotalShownDelta(self):
        if self.dynamic:
            return int(self.newTestsVisible)
        else:
            return 1  # we hide them temporarily for performance reasons, so can't do as above

    def tryAddTest(self, test, initial=False):
        iter = self.itermap.getIterator(test)
        if iter:
            self.addAdditional(iter, test)
            return iter
        suite = test.parent
        suiteIter = None
        if suite:
            suiteIter = self.tryAddTest(suite, initial)
        followIter = self.findFollowIter(suite, test, initial)
        return self.addSuiteWithParent(test, suiteIter, followIter)

    def findFollowIter(self, suite, test, initial):
        if not initial and suite:
            follower = suite.getFollower(test)
            if follower:
                return self.itermap.getIterator(follower)

    def addAdditional(self, iter, test):
        currTests = self.model.get_value(iter, 2)
        if not test in currTests:
            self.diag.info("Adding additional test to node " + self.model.get_value(iter, 0))
            currTests.append(test)

    def notifyRemove(self, test):
        delta = -test.size()
        iter = self.itermap.getIterator(test)
        allTests = self.model.get_value(iter, 2)
        if len(allTests) == 1:
            self.notify("TestTreeCounters", totalDelta=delta, totalShownDelta=delta, totalRowsDelta=delta)
            self.removeTest(test, iter)
        else:
            self.notify("TestTreeCounters", totalDelta=delta, totalShownDelta=delta, totalRowsDelta=0)
            allTests.remove(test)

    def removeTest(self, test, iter):
        self.diag.info("Removing test " + self.model.get_value(iter, 0))
        iterValid, filteredIter = self.findIter(test)
        self.selecting = True
        if self.selection.iter_is_selected(filteredIter):
            self.selection.unselect_iter(filteredIter)
        self.selecting = False
        self.selectionChanged(direct=False)
        self.model.remove(iter)
        self.itermap.remove(test)

    def notifyNameChange(self, test, origRelPath):
        iter = self.itermap.updateIterator(test, origRelPath)
        oldName = self.model.get_value(iter, 0)
        if test.name != oldName:
            self.model.set_value(iter, 0, test.name)

        iterValid, filteredIter = self.filteredModel.convert_child_iter_to_iter(iter)
        if self.selection.iter_is_selected(filteredIter):
            self.notify("NameChange", test, origRelPath)

    def notifyContentChange(self, *args):
        col_id, _ = self.model.get_sort_column_id()
        if col_id is None:
            self.model.set_default_sort_func(self.sortByTestCases)
        self.model.set_sort_column_id(Gtk.TREE_SORTABLE_DEFAULT_SORT_COLUMN_ID, Gtk.SortType.ASCENDING)

    def sortByTestCases(self, model, iter1, iter2, *args):
        test1 = self.model.get_value(iter1, 2)[0]
        test2 = self.model.get_value(iter2, 2)[0]
        index1 = test1.parent.testcases.index(test1)
        index2 = test2.parent.testcases.index(test2)
        return -1 if index1 < index2 else 1

    def notifyVisibility(self, tests, newValue):
        self.diag.info("Visibility change for " + repr(tests) + " to " + repr(newValue))
        if not newValue:
            self.selecting = True
        changedTests = []
        for test in tests:
            if self.updateVisibilityWithParents(test, newValue):
                changedTests.append(test)

        self.selecting = False
        if len(changedTests) > 0:
            self.diag.info("Actually changed tests " + repr(changedTests))
            self.notify("Visibility", changedTests, newValue)
            if self.treeView:
                self.updateVisibilityInViews(newValue)

    def updateVisibilityInViews(self, newValue):
        if newValue:  # if things have become visible, expand everything
            self.treeView.expand_all()
            GObject.idle_add(self.scrollToFirstTest)
        else:
            self.selectionChanged(direct=False)

    def updateVisibilityWithParents(self, test, newValue):
        changed = False
        if test.parent and newValue:
            changed |= self.updateVisibilityWithParents(test.parent, newValue)

        changed |= self.updateVisibilityInModel(test, newValue)
        if test.parent and not newValue and not self.hasVisibleChildren(test.parent) and not test.parent.state.hasStarted():
            self.diag.info("No visible children : hiding parent " + repr(test.parent))
            changed |= self.updateVisibilityWithParents(test.parent, newValue)
        return changed

    def isMarkedVisible(self, test):
        testIter = self.itermap.getIterator(test)
        # Can get None here when using queue systems, so that some tests in a suite
        # start processing when others have not yet notified the GUI that they have been read.
        return testIter is not None and self.model.get_value(testIter, 5) and test in self.model.get_value(testIter, 2)

    def updateVisibilityInModel(self, test, newValue):
        testIter = self.itermap.getIterator(test)
        if testIter is None:
            # Tests are not necessarily loaded yet in the GUI (for example if we do show only selected), don't stacktrace
            return False
        visibleTests = self.model.get_value(testIter, 2)
        isVisible = test in visibleTests
        changed = False
        if newValue and not isVisible:
            visibleTests.append(test)
            changed = True
        elif not newValue and isVisible:
            visibleTests.remove(test)
            changed = True

        if (newValue and len(visibleTests) > 1) or (not newValue and len(visibleTests) > 0):
            self.diag.info("No row visibility change : " + repr(test))
            return changed
        else:
            return self.setVisibility(testIter, newValue)

    def setVisibility(self, iter, newValue):
        oldValue = self.model.get_value(iter, 5)
        if oldValue == newValue:
            return False

        self.model.set_value(iter, 5, newValue)
        return True

    def hasVisibleChildren(self, suite):
        return any((self.isMarkedVisible(test) for test in suite.testcases))



#!/usr/bin/env python

from texttestlib import logconfiggen
import os
from copy import copy
from glob import glob


def generateForSelfTests(selftestDir, *args, **kw):
    if selftestDir:
        consoleGen = logconfiggen.PythonLoggingGenerator(os.path.join(
            selftestDir, "logging.console"), prefix="%(TEXTTEST_LOG_DIR)s/", **kw)
        enabledLoggerNames = stdInfo + [("storytext replay log", "stdout"), ("kill processes", "stdout")]
        consoleGen.generate(enabledLoggerNames, *args)

        staticGen = logconfiggen.PythonLoggingGenerator(os.path.join(
            selftestDir, "logging.static_gui"), prefix="%(TEXTTEST_LOG_DIR)s/", **kw)
        enabledLoggerNames = stdInfo + [("gui log", "gui_log"), ("storytext replay log", "gui_log")]
        staticGen.generate(enabledLoggerNames, *args)

        dynamicGen = logconfiggen.PythonLoggingGenerator(os.path.join(
            selftestDir, "logging.dynamic_gui"), prefix="%(TEXTTEST_LOG_DIR)s/", **kw)
        enabledLoggerNames = stdInfo + [("gui log", "dynamic_gui_log"), ("storytext replay log", "dynamic_gui_log"),
                                        ("kill processes", "dynamic_gui_log")]
        dynamicGen.generate(enabledLoggerNames, *args)


def getAppNames(logFileDir):
    rootDir = os.path.dirname(logFileDir)
    configFiles = glob(os.path.join(rootDir, "config.*"))
    return set([os.path.basename(f).split(".")[1] for f in configFiles])


def findSelfTestDirs():
    pattern = os.path.join(os.getenv("TEXTTEST_HOME"), "*", "log", "logging.console")
    files = glob(pattern)
    if len(files) == 0:
        logDir = os.path.join(os.getenv("TEXTTEST_HOME"), "log")
        return (logDir if os.path.isdir(logDir) else None), None

    selfTestDir, otherDir, otherAppName = None, None, None
    for f in files:
        d = os.path.dirname(f)
        appNames = getAppNames(d)
        if "texttest" in appNames:
            selfTestDir = d
        elif len(appNames) == 1:
            otherDir = d
            otherAppName = appNames.pop()
    return selfTestDir, otherDir, otherAppName


def combineLoggers(coreLoggers, storytextLoggers):
    allLoggers = copy(coreLoggers)
    debugLoggers = []
    for logger in storytextLoggers:
        if logger not in coreLoggers:
            allLoggers.append(logger)
            debugLoggers.append(logger)
    return allLoggers, debugLoggers


if __name__ == "__main__":
    consoleGen = logconfiggen.PythonLoggingGenerator("logging.console")
    stdInfo = [("standard log", "stdout")]
    killInfo = [("kill processes", "stdout")]
    consoleGen.generate(stdInfo + killInfo)

    batchGen = logconfiggen.PythonLoggingGenerator("logging.batch")
    batchGen.generate(stdInfo + killInfo, timeStdout=True)

    guiGen = logconfiggen.PythonLoggingGenerator("logging.gui")
    guiGen.generate(stdInfo, defaultLevel="WARNING")

    installationRoot = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    coreLoggers = logconfiggen.findLoggerNamesUnder(installationRoot)
    storytextLib = os.path.join(installationRoot, "../storytext", "lib")
    storytextLoggers = logconfiggen.findLoggerNamesUnder(storytextLib)
    allLoggers, debugLoggers = combineLoggers(coreLoggers, storytextLoggers)

    debugGen = logconfiggen.PythonLoggingGenerator("logging.debug", postfix="diag", prefix="%(TEXTTEST_PERSONAL_LOG)s/")
    debugGen.generate(enabledLoggerNames=[], allLoggerNames=allLoggers, debugLevelLoggers=debugLoggers)

    selfTestDir, siteSelfTestDir, siteAppName = findSelfTestDirs()

    generateForSelfTests(selfTestDir, allLoggers, debugLoggers, postfix="texttest")

    # Site-specific
    siteDiagFile = os.path.join(installationRoot, "../site/log/logging.debug")
    if os.path.isfile(siteDiagFile):
        siteLib = os.path.join(installationRoot, "../site", "lib")
        siteLoggers = logconfiggen.findLoggerNamesUnder(siteLib)
        allLoggers = sorted(allLoggers + siteLoggers)
        debugGen = logconfiggen.PythonLoggingGenerator(
            siteDiagFile, postfix="diag", prefix="%(TEXTTEST_PERSONAL_LOG)s/")
        debugGen.generate(enabledLoggerNames=[], allLoggerNames=allLoggers, debugLevelLoggers=debugLoggers)

        generateForSelfTests(siteSelfTestDir, allLoggers, debugLoggers, postfix=siteAppName)



# Plugin for bugzilla version 3.x that uses the new webservice interface. Tested for version 3.0.4
# but tried to make it a bit future-proof. The older "bugcli" interface is discontinued but there is
# a plugin for it too, see bugzillav2.py. Documentation can be found at e.g.
# http://www.astrogrid.org/bugzilla/docs/html/api/

# The webservice interface is still fairly primitive, there is for example no way to extract the initial
# comment. But at least it's officially part of bugzilla so will hopefully hang around for longer than
# bugcli did, and there are clearly plans to move it forward.

import xmlrpc.client
from collections import OrderedDict

def getEntry(dict, key):
    return dict.get(key, "UNKNOWN")


def filterInternals(internals, alreadyMentioned):
    accepted = []
    boringValues = [0, "", "---", "all", "All", "unspecified"] + alreadyMentioned
    for key, value in list(internals.items()):
        if key.find("accessible") == -1 and value not in boringValues:
            accepted.append((key.replace("_", " ").capitalize(), value))
    accepted.sort()
    return accepted


def isResolved(status):
    return status == "RESOLVED" or status == "CLOSED"


def parseReply(reply, location, id):
    try:
        bugInfo = reply["bugs"][0]
        # This is marked unstable: we won't rely on its contents containing anything in particular
        internals = bugInfo["internals"]
        bugId = getEntry(bugInfo, "id")
        summary = getEntry(bugInfo, "summary")
        status = getEntry(internals, "bug_status")
        internals = filterInternals(internals, [bugId, summary, status])
        ruler = "*" * 30 + "\n"
        message = ruler + "Summary: " + summary + "\nBug Status: " + status + "\n\n"
        for fieldName, value in internals:
            message += fieldName + ": "
            if type(value) == dict:
                message += str(OrderedDict(sorted(value.items())))
            else:
                message += str(value)
            message += "\n"
        message += ruler
        message += "\nView bug " + str(bugId) + " using bugzilla URL=" + location + \
            "/show_bug.cgi?id=" + str(bugId) + "\n"
        return status, message, isResolved(status), id
    except (IndexError, KeyError):
        message = ("Could not parse reply from bugzilla's web service, "
                   "maybe incompatible interface. Text of reply follows : \n"
                   + str(reply))
        return "BAD SCRIPT", message, False, id


def findBugInfo(bugId, location, *args):
    scriptLocation = location + "/xmlrpc.cgi"
    proxy = xmlrpc.client.ServerProxy(scriptLocation)
    try:
        return parseReply(proxy.Bug.get_bugs({"ids": [bugId]}), location, bugId)
    except xmlrpc.client.Fault as e:
        return "NONEXISTENT", e.faultString, False, bugId
    except Exception as e:
        message = ("Failed to communicate with '" + scriptLocation + "': "
                   + str(e)
                   + ".\n\nPlease make sure that the configuration entry "
                   "'bug_system_location' points to a correct location of a "
                   "Bugzilla version 3.x installation. The current value is '"
                   + location + "'.")
        return "BAD SCRIPT", message, False, bugId



#!/usr/bin/env python


from distutils.core import setup
from distutils.command.build_py import build_py
from distutils.command.install_scripts import install_scripts

import os
import shutil


class build_py_preserve_permissions(build_py):
    def copy_file(self, src, dst, preserve_mode=True, **kw):
        # Preserve mode of files under libexec
        if "libexec" in dst:
            return build_py.copy_file(self, src, dst, preserve_mode=True, **kw)
        else:
            return build_py.copy_file(self, src, dst, preserve_mode=preserve_mode, **kw)


# Lifted from bzr setup.py, use for Jython on Windows which has no native installer
class windows_install_scripts(install_scripts):
    """ Customized install_scripts distutils action.
    """

    def run(self):
        install_scripts.run(self)   # standard action
        src = os.path.join(self.install_dir, "texttest")
        dst = src + "c.py"
        if os.path.isfile(dst):
            os.remove(dst)
        os.rename(src, dst)
        with open(src + ".pyw", "w") as writeFile:
            with open(dst) as readFile:
                for line in readFile:
                    if line.startswith("#!"):
                        writeFile.write(line.replace("python.exe", "pythonw.exe"))
                    else:
                        writeFile.write(line)


command_classes = {"build_py": build_py_preserve_permissions}
if os.name == "nt":
    command_classes['install_scripts'] = windows_install_scripts

py_modules = []

packages = ["texttestlib", "texttestlib.default", "texttestlib.queuesystem",
            "texttestlib.default.batch", "texttestlib.default.gtkgui", "texttestlib.default.knownbugs",
            "texttestlib.default.gtkgui.default_gui", "texttestlib.default.gtkgui.version_control"]

package_data = {"texttestlib": ["doc/ChangeLog", "doc/quick_start.txt", "doc/CREDITS.txt", "doc/MigrationNotes*", "doc/LICENSE.txt",
                                "etc/*", "etc/.*", "libexec/*", "log/*", "images/*.*", "images/retro/*"],
                "texttestlib.default.batch": ["testoverview_javascript/*"]}
scripts = ["bin/texttest", "bin/filter_rundependent.py", "bin/filter_fpdiff.py"]
if os.name == "posix":
    scripts.append("texttestlib/libexec/interpretcore")

from texttestlib.texttest_version import version

setup(name='TextTest',
      version=version,
      author="Geoff Bache",
      author_email="geoff.bache@pobox.com",
      url="http://www.texttest.org",
      description="A tool for text-based Approval Testing",
      long_description="TextTest is a tool for text-based Approval Testing, which is an approach to acceptance testing/functional testing. In other words, it provides support for regression testing by means of comparing program output files against a specified approved versions of what they should look like.",
      packages=packages,
      package_dir={},
      py_modules=py_modules,
      package_data=package_data,
      classifiers=["Programming Language :: Python",
                   "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
                   "Operating System :: OS Independent",
                   "Development Status :: 5 - Production/Stable",
                   "Environment :: X11 Applications :: GTK",
                   "Environment :: Win32 (MS Windows)",
                   "Environment :: Console",
                   "Intended Audience :: Developers",
                   "Intended Audience :: Information Technology",
                   "Topic :: Software Development :: Testing",
                   "Topic :: Software Development :: Libraries :: Python Modules"],
      scripts=scripts,
      cmdclass=command_classes,
      install_requires=['psutil'],
      extras_require={
        'graphs': ['matplotlib']
      }

      )




import sys
import logging
import types
from texttestlib import plugins
from queue import Queue, Empty
from collections import OrderedDict
from threading import Lock

plugins.addCategory("cancelled", "cancelled", "were cancelled before starting")


class Cancelled(plugins.TestState):
    def __init__(self, briefText, freeText):
        plugins.TestState.__init__(self, "cancelled", briefText=briefText, freeText=freeText,
                                   started=1, completed=1, lifecycleChange="complete")

# We're set up for running in a thread but we don't do so by default, for simplicity


class BaseActionRunner(plugins.Responder, plugins.Observable):
    cancelFreeText = "Test run was cancelled before it had started"

    def __init__(self, optionMap, diag):
        plugins.Responder.__init__(self)
        plugins.Observable.__init__(self)
        self.optionMap = optionMap
        self.testQueue = Queue()
        self.exited = False
        self.allComplete = False
        self.lock = Lock()
        self.killSignal = None
        self.diag = diag
        self.lockDiag = logging.getLogger("locks")

    def notifyAdd(self, test, initial):
        if test.classId() == "test-case":
            self.diag.info("Adding test " + test.uniqueName)
            self.addTest(test)

    def addTest(self, test):
        self.testQueue.put(test)

    def notifyAllRead(self, *args):
        self.diag.info("All read, adding terminator")
        self.testQueue.put(None)

    def notifyComplete(self, test):
        if not self.exited and "stop" in self.optionMap and test.state.hasFailed():
            self.exited = True
            self.cancelFreeText = "Test run was cancelled due to previous failure of test " + test.getRelPath()

    def runAllTests(self):
        self.runQueue(self.getTestForRun, self.runTest, "running")
        self.cleanup()
        self.diag.info("Terminating")

    def notifyAllComplete(self):
        self.allComplete = True

    def notifyKill(self, test):
        self.lock.acquire()
        self.killOrCancel(test)
        self.lock.release()

    def notifyKillProcesses(self, sig=None):
        self.diag.info("Got exit!")
        if self.allComplete:
            return
        self.lockDiag.info("Trying to get lock for killing")
        self.lock.acquire()
        self.lockDiag.info("Got lock for killing")
        self.notify("Status", "Killing all running tests ...")
        self.killSignal = sig
        self.exited = True
        self.killTests()
        self.notify("Status", "Killed all running tests.")
        self.lock.release()

    def cancel(self, test, briefText="cancelled", freeText="", **kwargs):
        if not test.state.isComplete():
            if not freeText:
                freeText = self.cancelFreeText
            self.changeState(test, Cancelled(briefText, freeText), **kwargs)

    def changeState(self, test, state):
        test.changeState(state)  # for overriding in case we need other notifiers

    def runQueue(self, getMethod, runMethod, desc, block=True):
        while True:
            test = getMethod(block)
            if not test:  # completed normally
                break

            if self.exited:
                self.cancel(test)
                self.diag.info("Cancelled " + desc + " " + test.uniqueName)
            elif not test.state.isComplete():
                self.diag.info(desc.capitalize() + " test " + test.uniqueName)
                runMethod(test)
                self.diag.info("Completed " + desc + " " + test.uniqueName)

    def getItemFromQueue(self, queue, block, replaceTerminators=False):
        try:
            item = queue.get(block=block)
            if replaceTerminators and item is None:
                if not block:
                    item = self.getItemFromQueue(queue, False, True)
                queue.put(None)
            return item
        except Empty:
            return

    def getTestForRun(self, block=True):
        return self.getItemFromQueue(self.testQueue, block=block)

    def canBeMainThread(self):
        return False  # We block, so we shouldn't be the main thread...


class ActionRunner(BaseActionRunner):
    def __init__(self, optionMap, *args):
        BaseActionRunner.__init__(self, optionMap, logging.getLogger("Action Runner"))
        self.currentTestRunner = None
        self.previousTestRunner = None
        self.appRunners = OrderedDict()

    def addSuite(self, suite):
        plugins.log.info("Using " + suite.app.description(includeCheckout=True))
        appRunner = ApplicationRunner(suite, self.diag)
        self.appRunners[suite.app] = appRunner

    def notifyAllReadAndNotified(self):
        # kicks off processing. Don't use notifyAllRead as we end up running all the tests before
        # everyone's been notified of the reading.
        self.runAllTests()

    def notifyRerun(self, test):
        if self.currentTestRunner and self.currentTestRunner.test is test:
            self.diag.info("Got rerun notification for " + repr(test) + ", resetting actions")
            self.currentTestRunner.resetActionSequence()

    def runTest(self, test):
        # We have the lock coming in to here...
        appRunner = self.appRunners.get(test.app)
        if appRunner:
            self.lock.acquire()
            self.currentTestRunner = TestRunner(test, appRunner, self.diag, self.exited, self.killSignal)
            self.lock.release()

            self.currentTestRunner.performActions(self.previousTestRunner)
            self.previousTestRunner = self.currentTestRunner

            self.lock.acquire()
            self.currentTestRunner = None
            self.notifyComplete(test)
            self.lock.release()

    def killTests(self):
        if self.currentTestRunner:
            self.currentTestRunner.kill(self.killSignal)

    def killOrCancel(self, test):
        if self.currentTestRunner and self.currentTestRunner.test is test:
            self.currentTestRunner.kill()
        else:
            self.cancel(test)

    def getAllActionClasses(self):
        classes = set()
        for appRunner in list(self.appRunners.values()):
            for action in appRunner.actionSequence:
                classes.add(action.__class__)
        return classes

    def cleanup(self):
        for actionClass in self.getAllActionClasses():
            actionClass.finalise()
        for appRunner in list(self.appRunners.values()):
            appRunner.cleanActions()


class ActionsCompleteAction(plugins.Action):
    def __call__(self, test):
        test.actionsCompleted()

    def setUpSuite(self, suite):
        # Only occasionally needed, e.g. in ReplaceText script
        if suite.state.hasStarted():
            suite.actionsCompleted()

    def callDuringAbandon(self, *args):
        return True


class ApplicationRunner:
    def __init__(self, testSuite, diag):
        self.testSuite = testSuite
        self.suitesSetUp = {}
        self.suitesToSetUp = {}
        self.diag = diag
        self.actionSequence = self.getActionSequence()
        self.setUpApplications()

    def cleanActions(self):
        # clean up the actions before we exit
        self.suitesToSetUp = {}
        self.suitesSetUp = {}
        self.actionSequence = []

    def setUpApplications(self):
        for action in self.actionSequence:
            self.setUpApplicationFor(action)

    def setUpApplicationFor(self, action):
        self.diag.info("Performing " + str(action) + " set up on " + repr(self.testSuite.app))
        try:
            action.setUpApplication(self.testSuite.app)
        except Exception:
            sys.stderr.write("Exception thrown performing " + str(action) +
                             " set up on " + repr(self.testSuite.app) + " :\n")
            plugins.printException()

    def markForSetUp(self, suite):
        newActions = []
        for action in self.actionSequence:
            newActions.append(action)
        self.suitesToSetUp[suite] = newActions

    def setUpSuites(self, action, test):
        if test.parent:
            self.setUpSuites(action, test.parent)
        if test.classId() == "test-suite":
            if action in self.suitesToSetUp[test]:
                self.setUpSuite(action, test)
                self.suitesToSetUp[test].remove(action)

    def setUpSuite(self, action, suite):
        self.diag.info(str(action) + " set up " + repr(suite))
        action.setUpSuite(suite)
        if suite in self.suitesSetUp:
            self.suitesSetUp[suite].append(action)
        else:
            self.suitesSetUp[suite] = [action]

    def tearDownSuite(self, suite):
        self.diag.info("Try tear down " + repr(suite))
        actionsToTearDown = self.suitesSetUp.get(suite, [])
        for action in actionsToTearDown:
            self.diag.info(str(action) + " tear down " + repr(suite))
            action.tearDownSuite(suite)
        self.suitesSetUp[suite] = []

    def getActionSequence(self):
        actionSequenceFromConfig = self.testSuite.app.getActionSequence()
        actionSequence = []
        # Collapse lists and remove None actions
        for action in actionSequenceFromConfig:
            self.addActionToList(action, actionSequence)

        self.addActionToList(ActionsCompleteAction(), actionSequence)
        return actionSequence

    def addActionToList(self, action, actionSequence):
        if type(action) == list:
            for subAction in action:
                self.addActionToList(subAction, actionSequence)
        elif action != None:
            actionSequence.append(action)


class TestRunner:
    def __init__(self, test, appRunner, diag, killed, killSignal):
        self.test = test
        self.diag = diag
        self.actionSequence = []
        self.appRunner = appRunner
        self.killed = killed
        self.killSignal = killSignal
        self.currentAction = None
        self.lock = Lock()
        self.resetActionSequence()

    def resetActionSequence(self):
        self.setActionSequence(self.appRunner.actionSequence)

    def setActionSequence(self, actionSequence):
        self.actionSequence = []
        # Copy the action sequence, so we can edit it and mark progress
        for action in actionSequence:
            self.actionSequence.append(action)

    def handleExceptions(self, method, *args):
        try:
            method(*args)
            return True
        except plugins.TextTestError as e:
            self.failTest(str(e))
        except:
            exceptionText = plugins.getExceptionString()
            plugins.printWarning("Caught exception while running " + repr(self.test) +
                                 " changing state to UNRUNNABLE :\n" + exceptionText.rstrip(), stdout=True)
            self.failTest(exceptionText)
        return False

    def failTest(self, excString):
        execHosts = self.test.state.executionHosts
        failState = plugins.Unrunnable(freeText=excString, briefText="TEXTTEST EXCEPTION", executionHosts=execHosts)
        self.test.changeState(failState)

    def performActions(self, previousTestRunner):
        tearDownSuites, setUpSuites = self.findSuitesToChange(previousTestRunner)
        for suite in tearDownSuites:
            self.handleExceptions(previousTestRunner.appRunner.tearDownSuite, suite)
        for suite in setUpSuites:
            self.appRunner.markForSetUp(suite)
        abandon = self.test.state.shouldAbandon()
        while len(self.actionSequence):
            action = self.actionSequence.pop(0)
            if abandon and not action.callDuringAbandon(self.test):
                continue
            self.diag.info("->Performing action " + str(action) + " on " + repr(self.test))
            if self.handleExceptions(self.appRunner.setUpSuites, action, self.test):
                self.callAction(action)
            self.diag.info("<-End Performing action " + str(action))
            if not abandon and self.test.state.shouldAbandon():
                self.diag.info("Abandoning test...")
                abandon = True

    def kill(self, sig=None):
        self.diag.info("Killing test " + repr(self.test))
        self.lock.acquire()
        self.killed = True
        self.killSignal = sig
        if self.currentAction:
            self.currentAction.kill(self.test, sig)
        self.lock.release()

    def callAction(self, action):
        self.lock.acquire()
        self.currentAction = action
        if self.killed:
            action.kill(self.test, self.killSignal)
        self.lock.release()

        self.handleExceptions(action, self.test)

        self.lock.acquire()
        self.currentAction = None
        self.lock.release()

    def findSuitesToChange(self, previousTestRunner):
        tearDownSuites = []
        commonAncestor = None
        if previousTestRunner:
            commonAncestor = self.test.findCommonAncestor(previousTestRunner.test)
            self.diag.info("Common ancestor : " + repr(commonAncestor))
            tearDownSuites = previousTestRunner.findSuitesUpTo(commonAncestor)
        setUpSuites = self.findSuitesUpTo(commonAncestor)
        # We want to set up the earlier ones first
        setUpSuites.reverse()
        return tearDownSuites, setUpSuites

    def findSuitesUpTo(self, ancestor):
        suites = []
        currCheck = self.test.parent
        while currCheck != ancestor:
            suites.append(currCheck)
            currCheck = currCheck.parent
        return suites



import gi
from gi.repository import Gtk, GObject, GdkPixbuf
import os
import sys
import glob
from texttestlib import plugins, texttest_version
from .. import guiplugins, guiutils

# Show useful info about TextTest.
# I don't particularly like the standard Gtk.AboutDialog, and we also want
# to show pygtk/gtk/python versions in our dialog, so we create our own ...


class AboutTextTest(guiplugins.ActionResultDialogGUI):
    website = "http://www.texttest.org"

    def __init__(self, *args, **kw):
        self.creditsButton = None
        self.licenseButton = None
        self.versionsButton = None
        guiplugins.ActionResultDialogGUI.__init__(self, *args, **kw)

    def getDialogTitle(self):
        return "About TextTest"

    def isActiveOnCurrent(self, *args):
        return True

    def _getStockId(self):
        return "about"

    def _getTitle(self):
        return "_About TextTest"

    def messageAfterPerform(self):
        return ""

    def getTooltip(self):
        return "show information about texttest"

    def createButtons(self):
        self.creditsButton = self.dialog.add_button('texttest-stock-credits', Gtk.ResponseType.NONE)
        self.licenseButton = self.dialog.add_button('_License', Gtk.ResponseType.NONE)
        self.versionsButton = self.dialog.add_button('_Versions', Gtk.ResponseType.NONE)
        self.creditsButton.connect("clicked", self.showCredits)
        self.licenseButton.connect("clicked", self.showLicense)
        self.versionsButton.connect("clicked", self.showVersions)
        guiplugins.ActionResultDialogGUI.createButtons(self)

    def addContents(self):
        imageDir, retro = guiutils.getImageDir()
        imageType = "gif" if retro else "png"
        logoFile = os.path.join(imageDir, "texttest-logo." + imageType)
        logoPixbuf = GdkPixbuf.Pixbuf.new_from_file(logoFile)
        logo = Gtk.Image()
        logo.set_from_pixbuf(logoPixbuf)
        logoFrame = Gtk.Alignment.new(0.5, 0.5, 0.0, 0.0)
        logoFrame.set_padding(10, 10, 10, 10)
        logoFrame.add(logo)
        mainLabel = Gtk.Label()
        mainLabel.set_markup("<span size='xx-large'>TextTest " + texttest_version.version + "</span>\n")
        messageLabel = Gtk.Label()
        message = "TextTest is an application-independent tool for text-based\nfunctional testing. This means running a batch-mode program\nin lots of different ways, and using the text output produced\nas a means of controlling the behaviour of that application."
        messageLabel.set_markup("<i>" + message + "</i>\n")
        messageLabel.set_justify(Gtk.Justification.CENTER)
        # On Windows the default URI hook fails and causes trouble...
        # According to the docs you can set "None" here but that doesn't seem to work...
        # disabled this MB 2018-12-05
        # Gtk.link_button_set_uri_hook(lambda x, y : None)
        urlButton = Gtk.LinkButton(self.website, self.website)
        urlButton.set_property("border-width", 0)
        urlButtonbox = Gtk.HBox()
        urlButtonbox.pack_start(urlButton, True, False, 0)
        urlButton.connect("clicked", self.urlClicked)
        licenseLabel = Gtk.Label()
        licenseLabel.set_markup("<span size='small'>Copyright " + u'\xa9' + " The authors</span>\n")
        self.dialog.vbox.pack_start(logoFrame, False, False, 0)
        self.dialog.vbox.pack_start(mainLabel, True, True, 0)
        self.dialog.vbox.pack_start(messageLabel, False, False, 0)
        self.dialog.vbox.pack_start(urlButtonbox, False, False, 0)
        self.dialog.vbox.pack_start(licenseLabel, False, False, 0)
        self.dialog.set_resizable(False)

    def urlClicked(self, *args):
        status = guiplugins.openLinkInBrowser(self.website)
        self.notify("Status", status)

    def showCredits(self, *args):
        newDialog = TextFileDisplayDialog(self.validApps, False, {}, "CREDITS.txt", self.dialog)
        newDialog.performOnCurrent()

    def showLicense(self, *args):
        newDialog = TextFileDisplayDialog(self.validApps, False, {}, "LICENSE.txt", self.dialog)
        newDialog.performOnCurrent()

    def showVersions(self, *args):
        newDialog = ShowVersions(self.validApps)
        newDialog.performOnCurrent()


class ShowVersions(guiplugins.ActionResultDialogGUI):
    def isActiveOnCurrent(self, *args):
        return True

    def _getTitle(self):
        return "Component _Versions"

    def messageAfterPerform(self):
        return ""

    def getTooltip(self):
        return "show component version information"

    def getDialogTitle(self):
        return "Version Information"

    def makeString(self, versionTuple):
        if type(versionTuple) is str:
            return versionTuple
        else:
            return ".".join(map(str, versionTuple))

    def addTable(self, vbox, name, data, alignRight, columnSpacings, **kw):
        table = Gtk.Table(len(data), 2, homogeneous=False)
        table.set_row_spacings(1)
        table.set_col_spacings(columnSpacings)
        for rowNo, (title, versionTuple) in enumerate(data):
            table.attach(self.justify(title + ":", 0.0), 0, 1, rowNo, rowNo +
                         1, xoptions=Gtk.AttachOptions.FILL, xpadding=1)
            table.attach(self.justify(self.makeString(versionTuple), float(alignRight)), 1, 2, rowNo, rowNo + 1)

        header = Gtk.Label()
        header.set_markup("<b>You are using these " + name + "s:\n</b>")
        tableVbox = Gtk.VBox()
        tableVbox.pack_start(header, False, False, 0)
        tableVbox.pack_start(table, True, True, 0)
        centeredTable = Gtk.Alignment.new(0.5, 0.5, 1.0, 1.0)
        centeredTable.add(tableVbox)
        padding = 0
        if "padding" in kw:
            padding = kw["padding"]
            del kw["padding"]
        vbox.pack_start(centeredTable, True, True, padding, **kw)

    def addContents(self):
        versionList = [("TextTest", texttest_version.version),
                       ("Python", sys.version_info),
                       ("GTK", (Gtk.get_major_version(), Gtk.get_minor_version(), Gtk.get_micro_version())),
                       ("PyGI", gi.version_info)]

        installationList = [("Test Suite", os.getenv("TEXTTEST_HOME"))]
        if getattr(sys, 'frozen', False):
            installationList.append(("TextTest (frozen)", sys.executable))
        else:
            installationList += [("TextTest", plugins.installationRoots[0]),
                                 ("Python", sys.executable)]

        vbox = Gtk.VBox()
        self.addTable(vbox, "version", versionList, alignRight=True, columnSpacings=0)
        self.addTable(vbox, "installation", installationList, alignRight=False, columnSpacings=5, padding=10)
        frame = Gtk.Alignment.new(0.5, 0.5, 1.0, 1.0)
        frame.set_padding(10, 10, 10, 10)
        frame.add(vbox)
        self.dialog.vbox.pack_start(frame, True, True, 0)

    def justify(self, label, leftFill, markup=False):
        alignment = Gtk.Alignment.new(leftFill, 0.0, 0.0, 0.0)
        if markup:
            l = Gtk.Label()
            l.set_markup(label)
            alignment.add(l)
        else:
            alignment.add(Gtk.Label(label=label))
        return alignment


class TextFileDisplayDialog(guiplugins.ActionResultDialogGUI):
    def __init__(self, allApps, dynamic, inputOptions, fileName, parent=None):
        self.parent = parent
        self.fileName = fileName
        self.title = self.makeTitleFromFileName(fileName)
        guiplugins.ActionResultDialogGUI.__init__(self, allApps, dynamic)

    def getParentWindow(self):
        if self.parent:
            return self.parent
        else:
            return self.topWindow

    def makeTitleFromFileName(self, fileName):
        words = fileName.replace(".txt", "").split("_")
        return " ".join([word.capitalize() for word in words])

    def messageAfterPerform(self):
        return ""

    def _getTitle(self):
        return self.title

    def getTooltip(self):
        return "Show TextTest " + self.title

    def isActiveOnCurrent(self, *args):
        return True

    def getDialogTitle(self):
        return "TextTest " + self.title

    def getTabTitle(self):
        if self.title == "Credits":
            return "Written by"
        else:
            return self.title

    def addContents(self):
        try:
            file = open(plugins.installationPath("doc", self.fileName), encoding="utf8")
            text = file.read()
            file.close()
            buffer = Gtk.TextBuffer()
            buffer.set_text(text)
        except Exception as e:  # pragma : no cover - should never happen
            self.showErrorDialog("Failed to show " + self.fileName + " file:\n" + str(e))
            return

        textView = Gtk.TextView.new_with_buffer(buffer)
        textView.set_editable(False)
        textView.set_cursor_visible(False)
        textView.set_left_margin(5)
        textView.set_right_margin(5)
        useScrollbars = not self.parent and text.count("\n") > 30
        notebook = Gtk.Notebook()
        label = Gtk.Label(label=self.getTabTitle())
        if useScrollbars:
            window = Gtk.ScrolledWindow()
            window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
            window.add(textView)
            notebook.append_page(window, label)
            parentSize = self.topWindow.get_size()
            self.dialog.resize(int(parentSize[0] * 0.9), int(parentSize[1] * 0.7))
        else:
            notebook.append_page(textView, label)

        self.dialog.vbox.pack_start(notebook, True, True, 0)


class VersionInfoDialogGUI(guiplugins.ActionResultDialogGUI):
    def isActiveOnCurrent(self, *args):
        return True

    def messageAfterPerform(self):
        return ""

    def makeVersions(self, versionStr):
        versions = []
        for stringVersion in versionStr.split("."):
            try:
                versions.append(int(stringVersion))
            except ValueError:
                versions.append(stringVersion)
        return tuple(versions)

    def addContents(self):
        notebook = Gtk.Notebook()
        notebook.set_scrollable(True)
        notebook.popup_enable()
        docDir = plugins.installationDir("doc")
        versionInfo = self.readVersionInfo(docDir)
        for version in reversed(sorted(versionInfo.keys())):
            buffer = Gtk.TextBuffer()
            buffer.set_text(versionInfo[version])
            textView = Gtk.TextView.new_with_buffer(buffer)
            textView.set_editable(False)
            textView.set_cursor_visible(False)
            textView.set_left_margin(5)
            textView.set_right_margin(5)
            scrolledWindow = Gtk.ScrolledWindow()
            scrolledWindow.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
            scrolledWindow.add(textView)
            scrolledWindow.set_shadow_type(Gtk.ShadowType.IN)
            versionStr = ".".join(map(str, version))
            notebook.append_page(scrolledWindow, Gtk.Label(label=self.labelPrefix() + versionStr))

        if notebook.get_n_pages() == 0:  # pragma : no cover - should never happen
            raise plugins.TextTestError("\nNo " + self.getTitle() + " could be found in\n" + docDir + "\n")
        else:
            parentSize = self.topWindow.get_size()
            self.dialog.resize(int(parentSize[0] * 0.9), int(parentSize[1] * 0.7))
            self.dialog.vbox.pack_start(notebook, True, True, 0)

    def labelPrefix(self):
        return ""


class ShowMigrationNotes(VersionInfoDialogGUI):
    def _getTitle(self):
        return "_Migration Notes"

    def getTooltip(self):
        return "show texttest migration notes"

    def getDialogTitle(self):
        return "TextTest Migration Notes"

    def readVersionInfo(self, docDir):
        versionInfo = {}
        for fileName in glob.glob(os.path.join(docDir, "MigrationNotes*")):
            versions = self.makeVersions(fileName.split("_")[-1])
            versionInfo[versions] = open(fileName).read()
        return versionInfo

    def labelPrefix(self):
        return "from "


class ShowChangeLogs(VersionInfoDialogGUI):
    def _getTitle(self):
        return "_Change Logs"

    def getTooltip(self):
        return "show texttest change logs"

    def getDialogTitle(self):
        return "TextTest Change Logs"

    def readVersionInfo(self, docDir):
        versionInfo = {}
        currVersions = ()
        for line in open(os.path.join(docDir, "ChangeLog")):
            if line.startswith("Version"):
                words = line.strip().split()
                if len(words) == 2:
                    versionStr = words[-1].rstrip(":")
                    currVersions = self.makeVersions(versionStr)
                    versionInfo[currVersions] = ""
                else:
                    versionInfo[currVersions] += line
            elif len(line.strip()) == 0 or line.startswith("="):
                continue
            else:
                versionInfo[currVersions] += line
        return versionInfo


def getInteractiveActionClasses():
    classes = [ShowMigrationNotes, ShowChangeLogs, ShowVersions, AboutTextTest]
    for fileName in plugins.findDataPaths(["*.txt"], dataDirName="doc"):
        classes.append(plugins.Callable(TextFileDisplayDialog, os.path.basename(fileName)))
    return classes
